#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import asyncio
import inspect
import itertools
import logging
import os
import shlex
import traceback
from collections import defaultdict
from concurrent import futures
from contextlib import suppress
from datetime import datetime, timedelta
from functools import lru_cache, partial
from itertools import cycle
from uuid import UUID

from docopt import DocoptExit, docopt
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.eventloop import use_asyncio_event_loop
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout

from bluetooth_mesh.application import Application, Element
from bluetooth_mesh.apps import get_plugin_manager
from bluetooth_mesh.crypto import DeviceKey, NetworkKey
from bluetooth_mesh.messages.config import (
    GATTNamespaceDescriptor,
    PublishPeriodStepResolution,
)
from bluetooth_mesh.messages.properties import PropertyID
from bluetooth_mesh.messages.time import TimeRole, UNCERTAINTY_MS, CURRENT_TAI_UTC_DELTA
from bluetooth_mesh.models import (
    ConfigClient,
    ConfigServer,
    DebugClient,
    GatewayConfigClient,
    GenericLevelClient,
    GenericOnOffClient,
    HealthClient,
    LightCTLClient,
    LightExtendedControllerSetupClient,
    LightLightnessClient,
    NetworkDiagnosticClient,
    NetworkDiagnosticSetupClient,
    SceneClient,
    SensorClient,
)
from bluetooth_mesh.models.models import TimeClient

plugin_manager = get_plugin_manager()


class MeshCompleter(Completer):
    def __init__(self, application):
        self.application = application
        super().__init__()

    def get_completions(self, document, complete_event):
        if not document.is_cursor_at_the_end:
            return

        command = document.text_before_cursor
        if " " not in command:  # try commands
            for cmd in self.application.commands.values():
                if cmd.CMD.startswith(command):
                    yield Completion(cmd.CMD, start_position=-len(command))

        else:
            words = shlex.split(command)
            if command.endswith(" "):  # wo for difference in split and shlex.split
                words.append("")
            if command.startswith(LsCommand.CMD) or "-g" in words:  # try groups
                for group in self.application.network.groups:
                    group_name = group.name or ""

                    if group_name.startswith(words[-1]):
                        if " " in group_name:
                            group_name = '"{}"'.format(group_name)
                        yield Completion(group_name, start_position=-len(words[-1]))

            else:  # try nodes
                for node in {
                    node.uuid.hex[:4] for node in self.application.network.nodes
                }:
                    if node.startswith(words[-1]):
                        yield Completion(node, start_position=-len(words[-1]))


class Command:
    USAGE = """
    Usage:
        %(cmd)s
    """
    CMD = None

    async def __call__(self, application, arguments):
        raise NotImplementedError()

    def get_usage(self, **kwargs):
        return self.USAGE % dict(cmd=self.CMD, **kwargs)

    def __lt__(self, other):
        return self.CMD < other.CMD


class NodeSelectionCommandMixin:
    USAGE = """
    Usage:
        %(cmd)s <uuid>...
        %(cmd)s -g <groups>...

    Options:
        -g --groups

    Use <group>.<index> or <uuid>.<index> to talk to non-primary elements.
    """

    @staticmethod
    def get_addresses(application, arguments):
        uuids = {}
        for uuid_element in arguments.get("<uuid>", []):
            uuid, *element = uuid_element.split(".", maxsplit=1)
            uuids[uuid] = int(element[0]) if element else 0

        groups = {}
        for group_element in arguments.get("<groups>", []):
            group, *element = group_element.split(".", maxsplit=1)
            groups[group] = int(element[0]) if element else 0

        nodes = []
        for node in application.network.nodes:
            offset = uuids.get(node.uuid.hex[:4])
            if offset is not None:
                nodes.append(node.address + offset)

            offset = groups.get(application.network.get_node_group(node))
            if offset is not None:
                nodes.append(node.address + offset)

        return nodes


class LsCommand(Command):
    USAGE = """
    Usage:
        %(cmd)s [options] [<groups>...]

    Options:
        -l --long
    """

    CMD = "ls"

    async def __call__(self, application, arguments):
        groups = arguments["<groups>"]

        if not groups and arguments["--long"]:
            groups = [group.name for group in application.network.groups]

        if groups:
            nodes = defaultdict(list)

            for node in application.network.nodes:
                group = application.network.get_node_group(node)

                if not groups or group in groups:
                    nodes[group].append(node)

            for group, nodes in sorted(nodes.items(), key=lambda n: n[0]):
                yield "{}:".format(group)
                for node in sorted(nodes, key=lambda n: n.uuid):
                    if arguments["--long"]:
                        yield "\t{} {:04x} {}".format(
                            node.uuid, node.address, node.name
                        )
                    else:
                        yield "\t{}".format(node.name)
        else:
            for group in sorted(
                group.name for group in application.network.groups if group.name
            ):
                yield "\t{}".format(group)


class ModelCommandMixin:
    ELEMENT = None
    MODEL = None

    def get_model(self, application):
        return application.elements[self.ELEMENT][self.MODEL]


class AttentionCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -t --timeout=ATTENTION    Attention timer [default: 5].
        -g --groups
    """

    ELEMENT = 0
    MODEL = HealthClient
    CMD = "attention"

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        attention = int(arguments["--timeout"])

        tasks = []
        if arguments["<uuid>"]:
            tasks = [
                model.attention(node.address, app_index=0, attention=attention)
                for node in application.network.nodes
                if node.uuid.hex[:4] in arguments["<uuid>"]
            ]

        elif arguments["<groups>"]:
            tasks = [
                model.attention_unack(
                    application.network.get_group_address(self.MODEL, name=name),
                    app_index=0,
                    attention=attention,
                )
                for name in arguments["<groups>"]
            ]

        return asyncio.gather(*tasks)


class SceneCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s <uuid>...
        %(cmd)s -g <groups>...
        %(cmd)s -c <scene> [options] <uuid>...
        %(cmd)s -c <scene> [options] -g <groups>...

    Options:
        -c --scene=SCENE    Number of scene to recall.
        -t --transition-time=TRANSITION-TIME    Transition time in seconds [default: 0].
        -g --groups
    """
    ELEMENT = 0
    MODEL = SceneClient
    CMD = "scene"

    async def recall(self, model, application, arguments):
        scene_number = int(arguments["--scene"])
        transition_time = int(arguments["--transition-time"])

        tasks = []
        if arguments["<uuid>"]:
            tasks = [
                model.recall_scene_unack(
                    node.address,
                    app_index=0,
                    scene_number=scene_number,
                    transition_time=transition_time,
                )
                for node in application.network.nodes
                if node.uuid.hex[:4] in arguments["<uuid>"]
            ]

        elif arguments["<groups>"]:
            tasks = [
                model.recall_scene_unack(
                    application.network.get_group_address(self.MODEL, name=name),
                    app_index=0,
                    scene_number=scene_number,
                    transition_time=transition_time,
                )
                for name in arguments["<groups>"]
            ]

        for i in await asyncio.gather(*tasks):
            yield i

    async def get(self, model, application, arguments):
        addresses = self.get_addresses(application, arguments)
        results = await model.get_scene(nodes=addresses, app_index=0)

        for address, data in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)

            param = str(data["current_scene"]) if data is not None else None
            yield "{} | {}: {}".format(group, node.name, param)

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        method = self.get if arguments["--scene"] is None else self.recall

        async for i in method(model, application, arguments):
            yield i or ""


class ModelGetCommandMixin(ModelCommandMixin, NodeSelectionCommandMixin):
    PARAMETER = None

    def format(self, data):
        return str(data[self.PARAMETER])

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        get = getattr(model, "get_{}".format(self.PARAMETER))
        results = await get(addresses, 0)

        for address, data in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)

            param = self.format(data) if data is not None else None
            yield "{} | {}: {}".format(group, node.name, param)


class DebugCommand(ModelGetCommandMixin, Command):
    ELEMENT = 0
    MODEL = DebugClient


class UptimeCommand(DebugCommand):
    CMD = "uptime"
    PARAMETER = "uptime"

    def format(self, data):
        return str(timedelta(seconds=data["uptime"]))


class FaultCommand(DebugCommand):
    CMD = "fault"
    PARAMETER = "last_sw_fault"

    def format(self, data):
        return "{}, {}".format(data["fault"], timedelta(seconds=data["time"]))


class VersionCommand(DebugCommand):
    CMD = "version"
    PARAMETER = "firmware_version"

    def format(self, data):
        return str(data["version"])


class IvIndexCommand(DebugCommand):
    CMD = "ivindex"
    PARAMETER = "ivindex"


class ArapCommand(DebugCommand):
    CMD = "arap"
    PARAMETER = "arap_content"

    def format(self, data):
        return "\n" + "\n".join(
            "\t{:04x}: {:8d} ivi={:1d}".format(k, v["sequence"], v["ivi"])
            for k, v in data["nodes"].items()
        )


class StatsCommand(DebugCommand):
    CMD = "stats"
    PARAMETER = "system_stats"

    def format(self, data):
        return "\n" + "\n".join(
            "\t{:>10s}: {:5d}".format(k, v) for k, v in data["stats"].items()
        )


class AppVersionCommand(DebugCommand):
    CMD = "app_version"
    PARAMETER = "app_version"

    def format(self, data):
        return str(data["version"])


class ConfigCommand(ModelGetCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient


class TtlCommand(ConfigCommand):
    CMD = "ttl"
    PARAMETER = "default_ttl"

    def format(self, data):
        return data["TTL"]


class RelayCommand(ConfigCommand):
    CMD = "relay"
    PARAMETER = "relay"

    def format(self, data):
        return "%s <interval: %sms, count: %s>" % (
            data["relay"],
            data["retransmit"]["interval"],
            data["retransmit"]["count"],
        )


class PublicationCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "publication"
    USAGE = """
    Usage:
        %(cmd)s -e ELEMENT -m MODEL [options] <uuid>...
        %(cmd)s -e ELEMENT -m MODEL [options] -g <groups>...

    Options:
        -e --element=ELEMENT [default: 0]
        -a --address=ADDRESS
        -k --key-index=KEY_INDEX
        -t --ttl=TTL
        -r --resolution=RESOLUTION
        -s --steps=STEP
        -c --count=COUNT
        -i --interval=INTERVAL
        -m --model=MODEL

    Resolution is one of: 100ms 1s, 10s, 10m
    """

    def _get_resolution(self, arguments):
        RESOLUTIONS = {
            "100ms": PublishPeriodStepResolution.RESOLUTION_100_MS,
            "1s": PublishPeriodStepResolution.RESOLUTION_1_S,
            "10s": PublishPeriodStepResolution.RESOLUTION_10_S,
            "10m": PublishPeriodStepResolution.RESOLUTION_10_MIN,
        }
        resolution = arguments.get("--resolution", None)

        return RESOLUTIONS[resolution] if resolution is not None else None

    async def __call__(self, application, arguments):
        from bluetooth_mesh import models

        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        for address in addresses:
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)
            element_index = int(arguments["--element"])

            if arguments["--key-index"] is not None:
                interval = int(
                    timedelta(seconds=float(arguments["--interval"])).total_seconds()
                    * 1000
                )

                resolution = (
                    self._get_resolution(arguments)
                    or PublishPeriodStepResolution.RESOLUTION_10_S
                )
                steps = int(arguments.get("--steps", 6))

                status = await model.set_publication(
                    address,
                    0,
                    element_address=address + element_index,
                    publication_address=int(arguments["--address"], 16),
                    app_key_index=int(arguments["--key-index"]),
                    TTL=int(arguments["--ttl"]),
                    publish_step_resolution=resolution,
                    publish_number_of_steps=steps,
                    retransmit_count=int(arguments["--count"]),
                    retransmit_interval=interval,
                    model=getattr(models, arguments["--model"]),
                )
            else:
                status = await model.get_publication(
                    address,
                    0,
                    element_address=address + element_index,
                    model=getattr(models, arguments["--model"]),
                )

            yield "{} | {}: #{} {} addr={:04x}, ttl={} key=#{} period={}, retransmit={}/{}".format(
                group,
                node.name,
                element_index,
                status.model.__name__,
                status.publication_address,
                status.ttl,
                status.app_key_index,
                status.period.total_seconds(),
                status.retransmissions["count"],
                status.retransmissions["interval"].total_seconds(),
            )


class SubscribeCommand(Command):
    CMD = "subscribe"
    USAGE = """
    Usage:
        %(cmd)s -e ELEMENT -m MODEL -a ADDRESS

    Options:
        -e --element=ELEMENT
        -m --model=MODEL
        -a --address=ADDRESS
    """

    async def __call__(self, application: Application, arguments):
        from bluetooth_mesh import models

        model = application.get_model_instance(
            int(arguments["--element"]), getattr(models, arguments["--model"])
        )

        await model.subscribe(
            application.app_keys,
            int(arguments["--address"], 16),
            SubscribeCommand.on_message,
        )

    @staticmethod
    def on_message(source, destination, app_index, message):
        timestamp = datetime.now().strftime("%Y-%m-%d %T.%f")
        print(f"{timestamp} {source:04x} -> {destination:04x}: {message!r}")


class AddSubscriptionCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "add_subscription"
    USAGE = """
    Usage:
        %(cmd)s -a=ADDRESS -m MODEL [options] <uuid>...
        %(cmd)s -a=ADDRESS -m MODEL [options] -g <groups>...

    Options:
        -e --element=ELEMENT [default: 0]
        -a --address=ADDRESS
        -m --model=MODEL
    """

    async def __call__(self, application, arguments):
        from bluetooth_mesh import models

        model = self.get_model(application)
        destinations = self.get_addresses(application, arguments)

        element = int(arguments["--element"]) if arguments["--element"] is not None else 0
        address = int(arguments["--address"], 16) if arguments["--address"] is not None else 0
        if not destinations:
            yield "invalid uuid!"
        for dst in destinations:
            element_address = dst + element

            results = await model.add_subscription(
                dst, 0, element_address, address, model=getattr(models, arguments["--model"])
            )

            node = application.network.get_node(address=dst)
            if results is None:
                yield f"{dst} | Subscription Status not received"
            else:
                yield f"{dst} | {node.name}:\n{results}"


class BindAppKeyCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "bind_appkey"
    USAGE = """
    Usage:
        %(cmd)s -m MODEL [options] <uuid>...
        %(cmd)s -m MODEL [options] -g <groups>...

    Options:
        -e --element=ELEMENT [default: 0]
        -k --keyindex=KEY_INDEX [default: 0]
        -m --model=MODEL
    """

    async def __call__(self, application, arguments):
        from bluetooth_mesh import models

        model = self.get_model(application)
        destinations = self.get_addresses(application, arguments)
        if not destinations:
            yield "invalid uuid!"

        element = int(arguments["--element"]) if arguments["--element"] is not None else 0
        key_index = int(arguments["--keyindex"]) if arguments["--keyindex"] is not None else 0

        for dst in destinations:
            element_address = dst + element

            results = await model.bind_app_key(
                dst, 0, element_address, key_index, model=getattr(models, arguments["--model"]))

            node = application.network.get_node(address=dst)
            if results is None:
                yield f"{dst} | Bind AppKey Status not received"
            else:
                yield f"{dst} | {node.name}:\n{results}"


class UnsubscribeCommand(Command):
    CMD = "unsubscribe"
    USAGE = """
    Usage:
        %(cmd)s -e ELEMENT -m MODEL -a ADDRESS

    Options:
        -e --element=ELEMENT
        -m --model=MODEL
        -a --address=ADDRESS
    """

    async def __call__(self, application: Application, arguments):
        from bluetooth_mesh import models

        model = application.get_model_instance(
            int(arguments["--element"]), getattr(models, arguments["--model"])
        )

        await model.unsubscribe(
            int(arguments["--address"], 16), SubscribeCommand.on_message
        )


class GatewayConfigurationCommand(
    ModelCommandMixin, NodeSelectionCommandMixin, Command
):
    USAGE = (
        """
    Usage:
        %(cmd)s ("""
        "--mtu=MTU | --mac=MAC | --server=HOST:PORT | --reconnect=INTERVAL | "
        "--dns=IP | --ip=IP | --gateway=IP | --netmask=NETMASK | [--get] | "
        "--mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL | "
        "--mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP |"
        "--mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP --ip=IP --netmask=NETMASK --gateway=IP"
        """) <uuid>

        %(cmd)s --mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL <uuid>
        %(cmd)s --mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP <uuid>
        %(cmd)s --mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP --ip=IP --netmask=NETMASK --gateway=IP <uuid>
        %(cmd)s [--get=TYPE] <uuid>
    Options:
        --mtu=MTU               Set MTU size
        --mac=MAC               Set MAC address
        --server=SERVER:PORT    Set server's name nad port
        --reconnect=INTERVAL    Set reconnect interval
        --dns=IP                Set DNS IP address
        --ip=IP                 Set IP address
        --gateway=IP            Set gateway UP address
        --netmask=NETMASK       Set netmask
        --get                   Get current configuration
    """
    )

    CMD = "gateway"
    ELEMENT = 0
    MODEL = GatewayConfigClient

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        # strip'--' prefixes from options and drop Nones
        kwargs = {
            k.lstrip("-"): v
            for k, v in arguments.items()
            if k.startswith("--") and v is not None
        }

        kwargs.pop("get", None)

        if len(kwargs) > 1:
            method = partial(model.configuration_set, **self.parse_args(kwargs))
        elif len(kwargs) > 0:
            ((name, value),) = kwargs.items()
            method = partial(getattr(model, f"{name}_set"), **self.parse_args(kwargs))
        else:
            method = getattr(model, f"configuration_get")

        tasks = [
            method(destination=node.address, net_index=0)
            for node in application.network.nodes
            if node.uuid.hex[:4] in arguments["<uuid>"]
        ]

        results = await asyncio.gather(*tasks)
        self.parse_status_payload(results[0]["params"]["payload"])

    @staticmethod
    def parse_args(args):
        parsed = {}
        for key in args:
            if key in ("mtu", "reconnect", "netmask"):
                parsed[key] = int(args[key])
            if key in ("mac", "dns", "ip", "gateway"):
                parsed[key] = args[key]
            if key in "server":
                # convert HOST:PORT into a Tuple
                host, port = args[key].split(":")
                parsed[key] = (host, int(port))

        return parsed

    @staticmethod
    def parse_status_payload(payload):
        server_addr = payload["server_address"]

        if server_addr == "":
            server_addr = "NONE"

        parsed = (
            "revision: {}\n"
            "mac: {}\n"
            "mtu: {}\n"
            "ip: {}/{}\n"
            "gateway: {}\n"
            "dns: {}\n"
            "server: {}:{}\n"
            "reconnect_interval: {}\n"
            "dhcp: {}\n"
            "status_code: {}".format(
                payload["chip_revision_id"],
                payload["mac_address"],
                payload["mtu_size"],
                payload["ip_address"],
                payload["netmask"],
                payload["gateway_ip_address"],
                payload["dns_ip_address"],
                server_addr,
                payload["server_port_number"],
                payload["reconnect_interval"],
                str(payload["flags"]),
                str(payload["status_code"]),
            )
        )
        print(parsed)


class GatewayPacketsCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>

    Options:
        --clear     Clear packet's error counters
    """

    CMD = "packets"
    ELEMENT = 0
    MODEL = GatewayConfigClient

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        if arguments["--clear"]:
            method = partial(model.packets_clear)
        else:
            method = partial(model.packets_get)

        tasks = [
            method(node.address, net_index=0)
            for node in application.network.nodes
            if node.uuid.hex[:4] in arguments["<uuid>"]
        ]

        results = await asyncio.gather(*tasks)
        self.parse_status_packets(results[0]["params"]["payload"])

    @staticmethod
    def parse_status_packets(payload):
        parsed = (
            "rx_errors: {}\n"
            "tx_errors: {}\n"
            "bandwidth: {}\n"
            "connection_state: {}\n"
            "link_status: {}\n"
            "last_error: {}".format(
                payload["total_eth_rx_errors"],
                payload["total_eth_tx_errors"],
                payload["bandwidth"],
                str(payload["connection_state"]["conn_state"]),
                str(payload["connection_state"]["link_status"]),
                str(payload["connection_state"]["last_error"]),
            )
        )
        print(parsed)


class AclCommand(Command):
    USAGE = """
    Usage:
        %(cmd)s 
        %(cmd)s --grant <uuid> [--devkey <devkey> --netkey <netkey>]
        %(cmd)s --revoke <uuid>

    Options:
        -g --grant=<uuid>      Grant
        -d --devkey=<devkey>   Devkey
        -n --netkey=<netkey>   Netkey
        -r --revoke=<uuid>     Revoke token
    """
    CMD = "acl"

    def format(self, application):
        for uuid, token in application.token_ring.acl():
            try:
                name = application.network.get_node(uuid=uuid).name
            except KeyError:
                name = None
            yield uuid, name, token

    async def __call__(self, application, arguments):
        if arguments["--grant"]:
            uuid = UUID(arguments["--grant"])
            dev_key = arguments["--devkey"]
            net_key = arguments["--netkey"]

            if not all([dev_key, net_key]):
                try:
                    dev_key = application.network.get_node(uuid=uuid).device_key
                    net_key = application.network.network_keys[0]
                except KeyError:
                    logging.getLogger("acl_grant").warning(
                        "Some key data is missing "
                        "but node could not be found in network configuration to automagically download it."
                    )
                    return
            else:
                dev_key = DeviceKey(bytes.fromhex(dev_key))
                net_key = NetworkKey(bytes.fromhex(net_key))

            await application.acl_grant(uuid=uuid, dev_key=dev_key, net_key=net_key)

        if arguments["--revoke"]:
            await application.acl_revoke(uuid=UUID(hex=arguments["--revoke"]))

        for uuid, name, token in self.format(application):
            print("\t%s (%s): %s" % (uuid, name, token))
        print()


class LightCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
            Usage:
                %(cmd)s <uuid>... [--lightness <light>] [--temperature <temp>]
                %(cmd)s -g <groups>... [--lightness <light>] [--temperature <temp>]

            Options:
                -g --groups
                -l --lightness <light>
                -t --temperature <temp>
            """
    ELEMENT = 0
    CMD = "light"
    PARAMETER = "light_status"
    CTL_ELEMENT = 2

    def format(self, data_light, data_ctl):
        return "lightness {present_lightness}, temperature {present_ctl_temperature}".format(
            present_lightness=data_light.get("present_lightness")
            if data_light
            else None,
            present_ctl_temperature=data_ctl.get("present_ctl_temperature")
            if data_ctl
            else None,
        )

    async def __call__(self, application, arguments):
        addresses = self.get_addresses(application, arguments)

        self.MODEL = LightLightnessClient
        model = self.get_model(application)
        if arguments["--lightness"]:
            results_light = await model.set_lightness(
                nodes=addresses, app_index=0, lightness=int(arguments["--lightness"])
            )
        else:
            results_light = await model.get_lightness(nodes=addresses, app_index=0)

        self.MODEL = LightCTLClient
        model = self.get_model(application)
        if arguments["--temperature"]:
            results_ctl = await model.set_ctl(
                nodes=[addr + self.CTL_ELEMENT for addr in addresses],
                app_index=0,
                ctl_temperature=int(arguments["--temperature"]),
            )
        else:
            results_ctl = await model.get_ctl(
                nodes=[addr + self.CTL_ELEMENT for addr in addresses], app_index=0
            )

        for address in addresses:
            node = application.network.get_node(address=address)
            param = self.format(
                results_light[address], results_ctl[address + self.CTL_ELEMENT]
            )
            print("{}: {}".format(node.name, param))


class LightRangeCommand(ModelGetCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
        Usage:
            %(cmd)s <uuid>...
            %(cmd)s -g <groups>...
            %(cmd)s -m <max> <uuid>...
            %(cmd)s -m <max> -g <groups>...

        Options:
            -m --max=MAX Max lightness range
            -g --groups
        """
    ELEMENT = 0
    CMD = "light_range"
    MODEL = LightLightnessClient
    PARAMETER = "lightness_range"

    async def __call__(self, application, arguments):
        max_lightness_arg = arguments["--max"]
        if max_lightness_arg is None:
            async for i in super(LightRangeCommand, self).__call__(application, arguments):
                yield i
            return

        min_lightness = 256
        max_lightness = max(min_lightness, min(65535, int(max_lightness_arg)))

        model: LightLightnessClient = self.get_model(application)

        await asyncio.gather(
            model.set_lightness_range_unack(
                app_index=0,
                destination=address,
                min_lightness=min_lightness,
                max_lightness=max_lightness
            )
            for address in self.get_addresses(application, arguments)
        )

    def format(self, data):
        return "min={}, max={}".format(data["range_min"], data["range_max"])


class NetworkTransmissionCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
        Usage:
            %(cmd)s <uuid>... [--interval <millis>] [--count <count>]
            %(cmd)s -g <groups>... [--interval <millis>] [--count <count>]

        Options:
            -g --groups
            -s --interval <millis>
            -c --count <count>
        """
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "net_transmission"
    PARAMETER = "net_transmission"

    def format(self, data):
        return "interval={}ms, count={}".format(*data)

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        for address in addresses:
            if arguments["--interval"] and arguments["--count"]:
                results = await model.set_network_transmission(
                    address,
                    net_index=0,
                    interval=int(arguments["--interval"]),
                    count=int(arguments["--count"]),
                )
            else:
                results = await model.get_network_transmission(address, net_index=0)

            node = application.network.get_node(address=address)
            param = self.format(results) if results is not None else None
            print("{}: {}".format(node.name, param))


class CompositionDataCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "composition"
    PARAMETER = "composition_data"

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        for address in addresses:
            data = await model.get_composition_data([address], net_index=0)
            node = application.network.get_node(address=address)

            composition = data[address]["data"]

            yield "{}: CID {}, PID {}, VID {}, CRPL {}, Features {:016b}".format(
                node.name,
                composition["CID"],
                composition["PID"],
                composition["VID"],
                composition["CRPL"],
                composition["features"],
            )

            for i, ele in enumerate(composition["elements"]):
                yield "\tElement {}, location: {}".format(i, ele["location"])
                yield "\t\t   SIG Models: " + ", ".join(
                    "{:04x}".format(mod["model_id"]) for mod in ele["SIG_models"]
                )
                yield "\t\tVendor Models: " + ", ".join(
                    "{:04x}:{:04x}".format(mod["vendor_id"], mod["model_id"])
                    for mod in ele["vendor_models"]
                )


class GenericOnOffCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s <uuid>...
        %(cmd)s -g <groups>...

    Options:
        -g --groups
    """
    TARGET = None
    MODEL = GenericOnOffClient
    ELEMENT = 0
    PARAMETER = "set_onoff"

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        tasks = []
        if arguments["<uuid>"]:
            command = getattr(model, self.PARAMETER)
            tasks = [
                command(node.address, app_index=0, onoff=self.TARGET)
                for node in application.network.nodes
                if node.uuid.hex[:4] in arguments["<uuid>"]
            ]

        elif arguments["<groups>"]:
            command = getattr(model, "{}_unack".format(self.PARAMETER))
            tasks = [
                command(
                    application.network.get_group_address(self.MODEL, name=name),
                    app_index=0,
                    onoff=self.TARGET,
                )
                for name in arguments["<groups>"]
            ]

        await asyncio.gather(*tasks)


class GenericOnCommand(GenericOnOffCommand):
    CMD = "on"
    TARGET = True


class GenericOffCommand(GenericOnOffCommand):
    CMD = "off"
    TARGET = False


class MorseCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s -g <group> <text>

    Options:
        -g --group
    """
    CMD = "morse"
    MODEL = GenericLevelClient
    ELEMENT = 0
    PARAMETER = "set_onoff_unack"

    CODE = {
        "A": ".-",
        "B": "-...",
        "C": "-.-.",
        "D": "-..",
        "E": ".",
        "F": "..-.",
        "G": "--.",
        "H": "....",
        "I": "..",
        "J": ".---",
        "K": "-.-",
        "L": ".-..",
        "M": "--",
        "N": "-.",
        "O": "---",
        "P": ".--.",
        "Q": "--.-",
        "R": ".-.",
        "S": "...",
        "T": "-",
        "U": "..-",
        "V": "...-",
        "W": ".--",
        "X": "-..-",
        "Y": "-.--",
        "Z": "--..",
        "1": ".----",
        "2": "..---",
        "3": "...--",
        "4": "....-",
        "5": ".....",
        "6": "-....",
        "7": "--...",
        "8": "---..",
        "9": "----.",
        "0": "-----",
        ".": ".-.-.-",
        ",": "--..--",
        ":": "---...",
        "?": "..--..",
        "'": ".----.",
        "-": "-....-",
        "/": "-..-.",
        "@": ".--.-.",
        "=": "-...-",
    }

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        destination = application.network.get_group_address(
            self.MODEL, name=arguments["<group>"]
        )

        send_interval = 0.01
        retransmissions = 6

        set_unack = partial(
            model.set_level_unack,
            destination=destination,
            app_index=0,
            retransmissions=retransmissions,
            send_interval=send_interval,
        )

        on = partial(set_unack, level=32767)
        off = partial(set_unack, level=-24576)

        try:
            # warm up time
            await off()
            await asyncio.sleep(0.5)

            for letter in arguments["<text>"].upper():
                code = self.CODE[letter]
                logging.getLogger("morse").info("%s = %s", letter, code)

                for blink in code:
                    await on(delay=retransmissions * send_interval)

                    delay = (
                        retransmissions * send_interval + 0.1 if blink == "." else 0.5
                    )
                    await off(delay=delay)
                    await asyncio.sleep(delay + 0.1)

                await asyncio.sleep(0.3)
        finally:
            await set_unack(level=-32768)


class LightExtendedControllerCommand(
    ModelCommandMixin, NodeSelectionCommandMixin, Command
):
    USAGE = """
    Usage:
        %(cmd)s <uuid>...
        %(cmd)s -g <groups>...
    """
    ELEMENT = 0
    MODEL = LightExtendedControllerSetupClient
    CMD = "lec"

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        auto_resume = await model.get_auto_resume_mode(nodes=addresses, net_index=0)
        auto_timer = await model.get_auto_resume_timer(nodes=addresses, net_index=0)

        results = {
            k: (auto_resume.get(k), auto_timer.get(k))
            for k in {*auto_resume.keys(), *auto_timer.keys()}
        }

        for address, (resume, timer) in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)

            yield "{} | {}: resume={}, timer={}".format(group, node.name, resume, timer)


class SensorCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -d --descriptor
        -p --property=PROPERTY
    """
    CMD = "sensor"
    MODEL = SensorClient
    ELEMENT = 0

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        if arguments.get("--descriptor"):
            results = await model.get_descriptor(addresses, app_index=0)

            for address, descriptor in results.items():
                node = application.network.get_node(address=address)
                group = application.network.get_node_group(node)

                yield "{} | {}:\n{}".format(group, node.name, descriptor)

            return

        property = arguments.get("--property")
        if property:
            try:
                property_id = getattr(PropertyID, property)
            except AttributeError:
                property_id = PropertyID(int(property, 16))

            results = await model.get_sensor(
                addresses, app_index=0, property_id=property_id
            )

            for address, status in results.items():
                node = application.network.get_node(address=address)
                group = application.network.get_node_group(node)

                yield "{} | {}: {!r}".format(group, node.name, status)

            return

        raise DocoptExit


class TimeGetCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -g --groups

    """
    CMD = "time_get"
    MODEL = TimeClient
    ELEMENT = 0

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        results = await model.get_time(nodes=addresses, app_index=0)

        for address, pkt_time in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)
            if pkt_time is None:
                yield f"{group} | {node.name}\nTime Status not received"
            elif pkt_time["date"] is None:
                yield f"{group} | {node.name}\nTime is not set!"
            else:
                yield "{} | {}:\n{} +- {} s".format(
                    group,
                    node.name,
                    pkt_time["date"].strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                    pkt_time["uncertainty"].total_seconds(),
                )


class TimeCurrentSetCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -o --offset <seconds>

    """
    CMD = "time_set_current_time"
    MODEL = TimeClient
    ELEMENT = 0

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)
        offset_seconds = (
            int(arguments["--offset"]) if arguments["--offset"] is not None else 0
        )
        results = await model.set_time(
            addresses,
            0,
            date=(datetime.now().astimezone() + timedelta(seconds=offset_seconds)),
            uncertainty=timedelta(milliseconds=UNCERTAINTY_MS),
            tai_utc_delta=timedelta(seconds=CURRENT_TAI_UTC_DELTA),
            time_authority=True,
        )

        for address, pkt_time in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)
            if pkt_time is None:
                yield f"{group} | {node.name}\nTime Status not received"
            elif pkt_time["date"] is None:
                yield f"{group} | {node.name}\nTime is not set!"
            else:
                yield "{} | {}:\n{} +- {} s".format(
                    group,
                    node.name,
                    pkt_time["date"].strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                    pkt_time["uncertainty"].total_seconds(),
                )


class TimeRoleGetCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...
        %(cmd)s [options] --subnets <subnets>...

    """
    CMD = "time_role_get"
    MODEL = TimeClient
    ELEMENT = 0

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        results = await model.get_time_role(nodes=addresses, app_index=0)

        for address, pkt_time in results.items():
            if pkt_time is None:
                yield f"{group} | {node.name}:\nTime Role Status not received"
            else:
                node = application.network.get_node(address=address)
                group = application.network.get_node_group(node)

                yield f"{group} | {node.name}:\nTime Role: " + str(
                    TimeRole(pkt_time["time_role"])
                )


class TimeRoleSetCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -r --role <role>

    """
    CMD = "time_role_set"
    MODEL = TimeClient
    ELEMENT = 0

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)
        time_role = (
            TimeRole(int(arguments.get("--role")))
            if arguments.get("--role") is not None
            else TimeRole.TIME_CLIENT
        )

        results = await model.set_time_role(addresses, 0, time_role)
        for address, pkt_time in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)
            if pkt_time is None:
                yield f"{group} | {node.name}:\nTime Role Status not received"
            else:
                yield f"{group} | {node.name}:\nTime Role: " + str(
                    TimeRole(pkt_time["time_role"])
                )


class HelpCommand(Command):
    CMD = "help"

    async def __call__(self, application, arguments):
        for cmd in sorted(application.commands.values()):
            yield "\t{}".format(cmd.CMD)


class PrimaryElement(Element):
    LOCATION = GATTNamespaceDescriptor.FORTY_SECOND

    MODELS = [
        GenericLevelClient,
        LightLightnessClient,
        LightCTLClient,
        GatewayConfigClient,
        ConfigServer,
        ConfigClient,
        HealthClient,
        DebugClient,
        GenericOnOffClient,
        SceneClient,
        NetworkDiagnosticClient,
        NetworkDiagnosticSetupClient,
        LightExtendedControllerSetupClient,
        SensorClient,
        TimeClient,
    ]


application_mixins = itertools.chain(*get_plugin_manager().hook.application_mixins())


class MeshCommandLine(*application_mixins, Application):
    PATH = "/com/silvair/meshcli/v10"

    COMMANDS = [
        HelpCommand,
        LsCommand,
        AttentionCommand,
        SceneCommand,
        UptimeCommand,
        FaultCommand,
        VersionCommand,
        IvIndexCommand,
        ArapCommand,
        StatsCommand,
        AppVersionCommand,
        LightCommand,
        GenericOnCommand,
        GenericOffCommand,
        CompositionDataCommand,
        MorseCommand,
        NetworkTransmissionCommand,
        TtlCommand,
        RelayCommand,
        GatewayConfigurationCommand,
        GatewayPacketsCommand,
        AclCommand,
        PublicationCommand,
        SubscribeCommand,
        AddSubscriptionCommand,
        BindAppKeyCommand,
        UnsubscribeCommand,
        LightExtendedControllerCommand,
        LightRangeCommand,
        SensorCommand,
        TimeGetCommand,
        TimeCurrentSetCommand,
        TimeRoleGetCommand,
        TimeRoleSetCommand,
    ]

    COMPANY_ID = 0xFEE5
    PRODUCT_ID = 0x42
    VERSION_ID = 1
    CRPL = 0xF00

    ELEMENTS = {0: PrimaryElement}

    def __init__(self, loop: asyncio.AbstractEventLoop, arguments):
        self.config_dir = os.path.expanduser("~/.config/meshcli")
        os.makedirs(self.config_dir, exist_ok=True)

        super().__init__(loop)
        self.arguments = arguments

        self.history = FileHistory(f"{self.config_dir}/history")
        self.completer = MeshCompleter(self)
        self.session = PromptSession(
            history=self.history, completer=self.completer, complete_while_typing=False,
        )
        self.commands = {cmd.CMD: cmd() for cmd in self.COMMANDS}
        self._tid = 0

    def get_cache_args(self):
        return (f"{self.config_dir}/cache",), {}

    @property
    @lru_cache(maxsize=1)
    def dev_key(self) -> DeviceKey:
        try:
            with open(f"{self.config_dir}/{self.uuid}.key", "rb") as dev_key:
                return DeviceKey(dev_key.read())
        except FileNotFoundError:
            key = DeviceKey(os.urandom(16))

            with open(f"{self.config_dir}/{self.uuid}.key", "wb") as dev_key:
                dev_key.write(key.bytes)
            return key

    async def get_network(
        self, get_address=True, environment="preprod", partner_id="silvair"
    ):
        return await super().get_network(
            get_address, self.arguments["--env"], self.arguments["--partner"]
        )

    async def add_keys(self):
        for index, key in self.subnet_keys:
            await self.add_net_key(index, key)

        for index, bound, key in self.app_keys:
            await self.add_app_key(
                net_key_index=bound, app_key_index=index, app_key=key
            )

        for node in self.network.nodes:
            # don't overwrite my own key
            if node.address in range(self.address, self.address + len(self.ELEMENTS)):
                continue

            await self.management_interface.import_remote_node(
                node.address, len(node.elements), DeviceKey(node.device_key.bytes)
            )

        debug_client = self.get_model_instance(element=0, model=DebugClient)
        health_client = self.get_model_instance(element=0, model=HealthClient)
        time_client = self.get_model_instance(element=0, model=TimeClient)

        for index, *_ in self.app_keys:
            await debug_client.bind(index)
            await health_client.bind(index)
            await time_client.bind(index)

    async def run(self, commands):
        self.address, self.network, self.iv_index = await self.get_network()

        async with self:
            await self._run(commands)

    async def _run(self, commands):
        await self.connect(socket_path=f"{self.config_dir}/{self.uuid}.socket")
        await self.add_keys()
        self.logger.info(
            "Loaded network %s, %d nodes", self.network, len(self.network.nodes)
        )

        commands = commands or cycle([None])

        for line in commands:
            if line is None:
                line = await self.session.prompt("{}> ".format(self.uuid), async_=True)

                if not line.strip():
                    continue
            else:
                self.logger.info("Running command: %s", line)

            cmd, *argv = shlex.split(line)

            handler = self.commands.get(cmd, None)

            if not handler:
                print("Command not found: {}".format(cmd))
                continue

            usage = handler.get_usage()
            try:
                arguments = docopt(usage, argv, help=False)

                lines = []
                result = handler(self, arguments)

                if inspect.isasyncgen(result):
                    async for line in result:
                        lines.append(line)
                elif inspect.isawaitable(result):
                    result = await result

                    if result is not None:
                        lines.append(str())

                if lines:
                    print("\n".join(lines))

            except Exception as ex:
                print(traceback.format_exc())
            except futures.TimeoutError:
                print("Command timed out: {}".format(cmd))
            except KeyboardInterrupt:
                pass
            except DocoptExit:
                print(usage.strip("\n").rstrip())


def main():
    doc = """
        Mesh CLI

        Usage:
            meshcli [options] [<command>...]
            meshcli -h | --help | --version

        Options:
            -l --login <login>             User login to platform service (email)
            -p --password <password>       User password to platform service (!unsecured!)
            -n --project <project>         Project name or id to be loaded from platform

            --partner <partner>            Partner identifier [default: silvair]
            --env <environment>            Environment: dev,preprod or prod [default: preprod]

            --auth0-client-id <client_id>  Auth0 client id
            --auth0-secret <secret>        Auth0 secret (!unsecured!)
            --auth0-domain <domain>        Auth0 domain name
            --serial-number <serial_num>   Gateway serial number

            -d --debug
            -h --help                      Show this help message and exit
            --version                      Show version and exit
    """
    use_asyncio_event_loop()
    arguments = docopt(doc, version="stat_checker 0.5")

    logging.basicConfig(
        format="%(asctime)s %(name)-40s %(levelname)-8s %(filename)15s:%(lineno)3s  %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    loop = asyncio.get_event_loop()
    mesh_cli = MeshCommandLine(loop, arguments)

    with suppress(EOFError, KeyboardInterrupt), patch_stdout():
        loop.run_until_complete(mesh_cli.run(arguments.get("<command>")))



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    Const,
    Default,
    GreedyRange,
    Int8ul,
    Int16ul,
    Optional,
    Struct,
    Switch,
    this,
)

from bluetooth_mesh.messages.config import (
    TTL,
    ConfigHeartbeatHops,
    ConfigHeartbeatPublicationFeatures,
    Int12ul,
    UnicastUnassignedAddress,
    UnicastUnassignedGroupAddress,
)
from bluetooth_mesh.messages.generic.onoff import TransitionTime
from bluetooth_mesh.messages.util import EnumAdapter, Opcode

MAX_RECORD_COUNT = 32


class NetworkDiagnosticServerOpcode(IntEnum):
    SILVAIR_NDS = 0xFC3601


class NetworkDiagnosticServerSubOpcode(IntEnum):
    SUBSCRIPTION_GET = 0x00
    SUBSCRIPTION_SET = 0x01
    SUBSCRIPTION_SET_UNACK = 0x02
    SUBSCRIPTION_STATUS = 0x03
    RADIO_STAT_GET = 0x04
    RADIO_STAT_SET = 0x05
    RADIO_STAT_STATUS = 0x06


class NetworkDiagnosticSetupServerOpcode(IntEnum):
    SILVAIR_NDS_SETUP = 0xFD3601


class NetworkDiagnosticSetupServerSubOpcode(IntEnum):
    PUBLICATION_GET = 0x00
    PUBLICATION_SET = 0x01
    PUBLICATION_STATUS = 0x02


# fmt: off
RegistryRecord = Struct(
    "source" / UnicastUnassignedAddress,
    "count" / Int16ul,
    "min_hops" / ConfigHeartbeatHops,
    "max_hops" / ConfigHeartbeatHops
)

NetworkDiagnosticSetupServerPublicationSet = Struct(
    "destination" / UnicastUnassignedGroupAddress,
    "count" / Int16ul,
    "period" / TransitionTime,
    "ttl" / TTL,
    "net_key_index" / Int12ul,
    "features" / Optional(ConfigHeartbeatPublicationFeatures)
)

NetworkDiagnosticSetupServerPublicationStatus = NetworkDiagnosticSetupServerPublicationSet

NetworkDiagnosticServerSubscriptionSet = Struct(
    "destination" / UnicastUnassignedGroupAddress,
    "period" / Int16ul
)

NetworkDiagnosticServerSubscriptionStatus = Struct(
    "destination" / UnicastUnassignedGroupAddress,
    "period" / Int16ul,
    "max_record_count" / Const(MAX_RECORD_COUNT, Int8ul),
    "record" / GreedyRange(RegistryRecord)
)

NetworkDiagnosticServerPayload = Struct(
    "subopcode" / EnumAdapter(Int8ul, NetworkDiagnosticServerSubOpcode),
    "payload" / Default(Switch(
        this.subopcode,
        {
            NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_SET: NetworkDiagnosticServerSubscriptionSet,
            NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_SET_UNACK: NetworkDiagnosticServerSubscriptionSet,
            NetworkDiagnosticServerSubOpcode.SUBSCRIPTION_STATUS: NetworkDiagnosticServerSubscriptionStatus
        }
    ), None)
)

NetworkDiagnosticSetupServerPayload = Struct(
    "subopcode" / EnumAdapter(Int8ul, NetworkDiagnosticSetupServerSubOpcode),
    "payload" / Default(Switch(
        this.subopcode,
        {
            NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_SET: NetworkDiagnosticSetupServerPublicationSet,
            NetworkDiagnosticSetupServerSubOpcode.PUBLICATION_STATUS: NetworkDiagnosticSetupServerPublicationStatus,
        }
    ), None)
)

NetworkDiagnosticServerMessage = Struct(
    "opcode" / Const(NetworkDiagnosticServerOpcode.SILVAIR_NDS, Opcode(NetworkDiagnosticServerOpcode)),
    "params" / NetworkDiagnosticServerPayload
)

NetworkDiagnosticSetupServerMessage = Struct(
    "opcode" / Const(NetworkDiagnosticSetupServerOpcode.SILVAIR_NDS_SETUP, Opcode(NetworkDiagnosticSetupServerOpcode)),
    "params" / NetworkDiagnosticSetupServerPayload
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0223

from datetime import datetime, timedelta
from enum import IntEnum
from math import log, pow

from construct import (
    Adapter,
    Array,
    BitsInteger,
    Byte,
    BytesInteger,
    Embedded,
    Flag,
    Int8sl,
    Int8ul,
    Int16sl,
    Int16ul,
    Int24ul,
    Int32ul,
    PaddedString,
    Probe,
    Struct,
    Switch,
    this,
)

from bluetooth_mesh.messages.config import EmbeddedBitStruct
from bluetooth_mesh.messages.util import DefaultCountValidator


class PropertyID(IntEnum):
    AVERAGE_AMBIENT_TEMPERATURE_IN_A_PERIOD_OF_DAY = 0x0001
    AVERAGE_INPUT_CURRENT = 0x0002
    AVERAGE_INPUT_VOLTAGE = 0x0003
    AVERAGE_OUTPUT_CURRENT = 0x0004
    AVERAGE_OUTPUT_VOLTAGE = 0x0005
    CENTER_BEAM_INTENSITY_AT_FULL_POWER = 0x0006
    CHROMATICITY_TOLERANCE = 0x0007
    COLOR_RENDERING_INDEX_R9 = 0x0008
    COLOR_RENDERING_INDEX_RA = 0x0009
    DEVICE_APPEARANCE = 0x000A
    DEVICE_COUNTRY_OF_ORIGIN = 0x000B
    DEVICE_DATE_OF_MANUFACTURE = 0x000C
    DEVICE_ENERGY_USE_SINCE_TURN_ON = 0x000D
    DEVICE_FIRMWARE_REVISION = 0x000E
    DEVICE_GLOBAL_TRADE_ITEM_NUMBER = 0x000F
    DEVICE_HARDWARE_REVISION = 0x0010
    DEVICE_MANUFACTURER_NAME = 0x0011
    DEVICE_MODEL_NUMBER = 0x0012
    DEVICE_OPERATING_TEMPERATURE_RANGE_SPECIFICATION = 0x0013
    DEVICE_OPERATING_TEMPERATURE_STATISTICAL_VALUES = 0x0014
    DEVICE_OVER_TEMPERATURE_EVENT_STATISTICS = 0x0015
    DEVICE_POWER_RANGE_SPECIFICATION = 0x0016
    DEVICE_RUNTIME_SINCE_TURN_ON = 0x0017
    DEVICE_RUNTIME_WARRANTY = 0x0018
    DEVICE_SERIAL_NUMBER = 0x0019
    DEVICE_SOFTWARE_REVISION = 0x001A
    DEVICE_UNDER_TEMPERATURE_EVENT_STATISTICS = 0x001B
    INDOOR_AMBIENT_TEMPERATURE_STATISTICAL_VALUES = 0x001C
    INITIAL_CIE_1931_CHROMATICITY_COORDINATES = 0x001D
    INITIAL_CORRELATED_COLOR_TEMPERATURE = 0x001E
    INITIAL_LUMINOUS_FLUX = 0x001F
    INITIAL_PLANCKIAN_DISTANCE = 0x0020
    INPUT_CURRENT_RANGE_SPECIFICATION = 0x0021
    INPUT_CURRENT_STATISTICS = 0x0022
    INPUT_OVER_CURRENT_EVENT_STATISTICS = 0x0023
    INPUT_OVER_RIPPLE_VOLTAGE_EVENT_STATISTICS = 0x0024
    INPUT_OVER_VOLTAGE_EVENT_STATISTICS = 0x0025
    INPUT_UNDER_CURRENT_EVENT_STATISTICS = 0x0026
    INPUT_UNDER_VOLTAGE_EVENT_STATISTICS = 0x0027
    INPUT_VOLTAGE_RANGE_SPECIFICATION = 0x0028
    INPUT_VOLTAGE_RIPPLE_SPECIFICATION = 0x0029
    INPUT_VOLTAGE_STATISTICS = 0x002A
    LIGHT_CONTROL_AMBIENT_LUXLEVEL_ON = 0x002B
    LIGHT_CONTROL_AMBIENT_LUXLEVEL_PROLONG = 0x002C
    LIGHT_CONTROL_AMBIENT_LUXLEVEL_STANDBY = 0x002D
    LIGHT_CONTROL_LIGHTNESS_ON = 0x002E
    LIGHT_CONTROL_LIGHTNESS_PROLONG = 0x002F
    LIGHT_CONTROL_LIGHTNESS_STANDBY = 0x0030
    LIGHT_CONTROL_REGULATOR_ACCURACY = 0x0031
    LIGHT_CONTROL_REGULATOR_KID = 0x0032
    LIGHT_CONTROL_REGULATOR_KIU = 0x0033
    LIGHT_CONTROL_REGULATOR_KPD = 0x0034
    LIGHT_CONTROL_REGULATOR_KPU = 0x0035
    LIGHT_CONTROL_TIME_FADE = 0x0036
    LIGHT_CONTROL_TIME_FADE_ON = 0x0037
    LIGHT_CONTROL_TIME_FADE_STANDBY_AUTO = 0x0038
    LIGHT_CONTROL_TIME_FADE_STANDBY_MANUAL = 0x0039
    LIGHT_CONTROL_TIME_OCCUPANCY_DELAY = 0x003A
    LIGHT_CONTROL_TIME_PROLONG = 0x003B
    LIGHT_CONTROL_TIME_RUN_ON = 0x003C
    LUMEN_MAINTENANCE_FACTOR = 0x003D
    LUMINOUS_EFFICACY = 0x003E
    LUMINOUS_ENERGY_SINCE_TURN_ON = 0x003F
    LUMINOUS_EXPOSURE = 0x0040
    LUMINOUS_FLUX_RANGE = 0x0041
    MOTION_SENSED = 0x0042
    MOTION_THRESHOLD = 0x0043
    OPEN_CIRCUIT_EVENT_STATISTICS = 0x0044
    OUTDOOR_STATISTICAL_VALUES = 0x0045
    OUTPUT_CURRENT_RANGE = 0x0046
    OUTPUT_CURRENT_STATISTICS = 0x0047
    OUTPUT_RIPPLE_VOLTAGE_SPECIFICATION = 0x0048
    OUTPUT_VOLTAGE_RANGE = 0x0049
    OUTPUT_VOLTAGE_STATISTICS = 0x004A
    OVER_OUTPUT_RIPPLE_VOLTAGE_EVENT_STATISTICS = 0x004B
    PEOPLE_COUNT = 0x004C
    PRESENCE_DETECTED = 0x004D
    PRESENT_AMBIENT_LIGHT_LEVEL = 0x004E
    PRESENT_AMBIENT_TEMPERATURE = 0x004F
    PRESENT_CIE_1931_CHROMATICITY_COORDINATES = 0x0050
    PRESENT_CORRELATED_COLOR_TEMPERATURE = 0x0051
    PRESENT_DEVICE_INPUT_POWER = 0x0052
    PRESENT_DEVICE_OPERATING_EFFICIENCY = 0x0053
    PRESENT_DEVICE_OPERATING_TEMPERATURE = 0x0054
    PRESENT_ILLUMINANCE = 0x0055
    PRESENT_INDOOR_AMBIENT_TEMPERATURE = 0x0056
    PRESENT_INPUT_CURRENT = 0x0057
    PRESENT_INPUT_RIPPLE_VOLTAGE = 0x0058
    PRESENT_INPUT_VOLTAGE = 0x0059
    PRESENT_LUMINOUS_FLUX = 0x005A
    PRESENT_OUTDOOR_AMBIENT_TEMPERATURE = 0x005B
    PRESENT_OUTPUT_CURRENT = 0x005C
    PRESENT_OUTPUT_VOLTAGE = 0x005D
    PRESENT_PLANCKIAN_DISTANCE = 0x005E
    PRESENT_RELATIVE_OUTPUT_RIPPLE_VOLTAGE = 0x005F
    RELATIVE_DEVICE_ENERGY_USE_IN_A_PERIOD_OF_DAY = 0x0060
    RELATIVE_DEVICE_RUNTIME_IN_A_GENERIC_LEVEL_RANGE = 0x0061
    RELATIVE_EXPOSURE_TIME_IN_AN_ILLUMINANCE_RANGE = 0x0062
    RELATIVE_RUNTIME_IN_A_CORRELATED_COLOR_TEMPERATURE_RANGE = 0x0063
    RELATIVE_RUNTIME_IN_A_DEVICE_OPERATING_TEMPERATURE_RANGE = 0x0064
    RELATIVE_RUNTIME_IN_AN_INPUT_CURRENT_RANGE = 0x0065
    RELATIVE_RUNTIME_IN_AN_INPUT_VOLTAGE_RANGE = 0x0066
    SHORT_CIRCUIT_EVENT_STATISTICS = 0x0067
    TIME_SINCE_MOTION_SENSED = 0x0068
    TIME_SINCE_PRESENCE_DETECTED = 0x0069
    TOTAL_DEVICE_ENERGY_USE = 0x006A
    TOTAL_DEVICE_OFF_ON_CYCLES = 0x006B
    TOTAL_DEVICE_POWER_ON_CYCLES = 0x006C
    TOTAL_DEVICE_POWER_ON_TIME = 0x006D
    TOTAL_DEVICE_RUNTIME = 0x006E
    TOTAL_LIGHT_EXPOSURE_TIME = 0x006F
    TOTAL_LUMINOUS_ENERGY = 0x0070
    PRECISE_TOTAL_DEVICE_ENERGY_USE = 0x0072

    def __repr__(self):
        return str(self.value)


class TimeExponential8Validator(Adapter):
    def _decode(self, obj, content, path):
        return round(pow(1.1, obj - 64), 4) if obj else 0

    def _encode(self, obj, content, path):
        return round(log(obj, 1.1)) + 64 if obj else 0


class DateValidator(Adapter):
    EPOCH = datetime(1970, 1, 1)

    def _decode(self, obj, content, path):
        return None if obj == 0x0 else self.EPOCH + timedelta(days=obj)

    def _encode(self, obj, content, path):
        return 0x0 if obj is None else (obj - self.EPOCH).days


def FixedString(size):
    return PaddedString(size, "utf8")


# fmt: off


# time
TimeMiliseconds24 = Struct(
    "seconds" / DefaultCountValidator(Int24ul, rounding=3, resolution=0.001)
)

TimeHour24 = Struct(
    "hours" / DefaultCountValidator(Int24ul)
)

TimeSecond16 = Struct(
    "seconds" / DefaultCountValidator(Int16ul)
)

TimeExponential8 = Struct(
    "seconds" / TimeExponential8Validator(Int8ul)
)

TimeDecihour8 = Struct(
    "hour" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.1)
)

DateUTC = Struct(
    "date" / DateValidator(Int24ul)
)


# electric current
ElectricCurrent = Struct(
    "current" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01)
)

AverageCurrent = Struct(
    "electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    "sensing_duration" / TimeExponential8,
)

ElectricCurrentRange = Struct(
    "minimum_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    "maximum_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
)

ElectricCurrentSpecification = Struct(
    "minimum_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    "typical_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    "maximum_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
)

ElectricCurrentStatistics = Struct(
    "average_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    "standard_deviation_electric_current_value" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    Embedded(ElectricCurrentRange),
    "sensing_duration" / TimeExponential8,
)

RelativeValueInACurrentRange = Struct(
    "relative_value" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.5),
    "minimum_current" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01),
    "maximum_current" / DefaultCountValidator(Int16ul, rounding=2, resolution=0.01)
)


# voltage
Voltage = Struct(
    "voltage" / DefaultCountValidator(Int16ul, resolution=1/64),
)

AverageVoltage = Struct(
    "voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "sensing_duration" / TimeExponential8,
    Probe(this)
)

VoltageRange = Struct(
    "minimum_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "typical_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "maximum_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
)

VoltageStatistics = Struct(
    "average_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "standard_deviation_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "minimum_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "maximum_voltage_value" / DefaultCountValidator(Int16ul, resolution=1/64),
    "sensing_duration" / TimeExponential8,
)

RelativeValueInAVoltageRange = Struct(
    "relative_value" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.5),
    "minimum_voltage" / DefaultCountValidator(Int16ul, resolution=1/64),
    "maximum_voltage" / DefaultCountValidator(Int16ul, resolution=1/64),
)


# energy
Energy = Struct(
    "energy" / DefaultCountValidator(Int24ul)
)

PreciseEnergy = Struct(
    "energy" / DefaultCountValidator(Int32ul)
)


EnergyInAPeriodOfDay = Struct(
    "energy_value" / DefaultCountValidator(Int24ul),
    "start_time" / TimeDecihour8,
    "end_time" / TimeDecihour8,
)


# power
Power = Struct(
    "power" / DefaultCountValidator(Int24ul, rounding=1, resolution=0.1)
)

PowerSpecification = Struct(
    "minimum_power_value" / DefaultCountValidator(Int24ul, rounding=1, resolution=0.1),
    "typical_power_value" / DefaultCountValidator(Int24ul, rounding=1, resolution=0.1),
    "maximum_power_value" / DefaultCountValidator(Int24ul, rounding=1, resolution=0.1)
)


# temperature
Temperature = Struct(
    "temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01)
)

Temperature8 = Struct(
    "temperature" / DefaultCountValidator(Int8sl, rounding=1, resolution=0.5)
)

TemperatureRange = Struct(
    "minimum_temperature" / DefaultCountValidator(Int8sl, rounding=1, resolution=0.5),
    "maximum_temperature" / DefaultCountValidator(Int8sl, rounding=1, resolution=0.5),
)

Temperature8Statistics = Struct(
    "average_temperature" / DefaultCountValidator(Int8sl, rounding=1, resolution=0.5),
    "standard_deviation_temperature" / DefaultCountValidator(Int8sl, rounding=1, resolution=0.5),
    Embedded(TemperatureRange),
    "sensing_duration" / TimeExponential8,
)

Temperature8InAPeriodOfDay = Struct(
    "temperature" / DefaultCountValidator(Int8sl, rounding=1, resolution=0.5),
    "start_time" / TimeDecihour8,
    "end_time" / TimeDecihour8,
)

TemperatureStatistics = Struct(
    "average_temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01),
    "standard_deviation_temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01),
    "minimum_temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01),
    "maximum_temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01),
    "sensing_duration" / TimeExponential8,
)

RelativeValueInATemperatureRange = Struct(
    "relative_value" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.5),
    "minimum_temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01),
    "maximum_temperature" / DefaultCountValidator(Int16sl, rounding=2, resolution=0.01)
)


# luminosity
LuminousFlux = Struct(
    "luminous_flux" / DefaultCountValidator(Int16ul)
)

LuminousFluxRange = Struct(
    "minimum_luminous_flux" / DefaultCountValidator(Int16ul),
    "maximum_luminous_flux" / DefaultCountValidator(Int16ul)
)

LuminousEnergy = Struct(
    "luminous_energy" / DefaultCountValidator(Int24ul, rounding=1, resolution=1000),
)

LuminousExposure = Struct(
    "luminous_exposure" / DefaultCountValidator(Int24ul, rounding=1, resolution=1000),
)

LuminousIntensity = Struct(
    "luminous_intensity" / DefaultCountValidator(Int16ul)
)

LuminousEfficacy = Struct(
    "luminous_efficacy" / DefaultCountValidator(Int16ul, rounding=1, resolution=0.1)
)

Illuminance = Struct(
    "illuminance" / DefaultCountValidator(Int24ul, rounding=2, resolution=0.01)
)

RelativeValueInAnIlluminanceRange = Struct(
    "relative_value" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.5),
    "minimum_illuminance" / DefaultCountValidator(Int24ul, rounding=2, resolution=0.01),
    "maximum_illuminance" / DefaultCountValidator(Int24ul, rounding=2, resolution=0.01)
)

PerceivedLightness = Struct(
    "perceived_lightness" / Int16ul
)


# counters
Percentage8 = Struct(
    "percentage" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.5)
)

Count16 = Struct(
    "count" / DefaultCountValidator(Int16ul)
)

Count24 = Struct(
    "count" / DefaultCountValidator(Int24ul)
)

Coefficient = Struct(
    "coefficient" / DefaultCountValidator(Int32ul)
)


# chromaticity
ChromaticityTolerance = Struct(
    "chromaticity_tolerance" / DefaultCountValidator(Int8sl, rounding=4, resolution=0.0001)
)

ChromaticDistanceFromPlanckian = Struct(
    "distance_from_planckian" / DefaultCountValidator(Int16sl, rounding=5, resolution=0.00001)
)

CorrelatedColorTemperature = Struct(
    "correlated_color_temperature" / DefaultCountValidator(Int16ul, rounding=1, resolution=1)
)

ChromaticityCoordinates = Struct(
    "chromaticity_x_coordinate" / DefaultCountValidator(Int16ul, resolution=1/0xffff),
    "chromaticity_y_coordinate" / DefaultCountValidator(Int16ul, resolution=1/0xffff)
)

ColorRenderingIndex = Struct(
    "color_rendering_index" / DefaultCountValidator(Int8sl)
)


# misc
GlobalTradeItemNumber = Struct(
    "global_trade_item_number" / BytesInteger(6, swapped=True)
)

Appearance = Struct(  # TODO: check if correct
    *EmbeddedBitStruct(
        "_",
        "category" / BitsInteger(10),
        "sub_category" / BitsInteger(6),
        reversed=True
    )
)

CountryCode = Struct(
    "country_code" / DefaultCountValidator(Int16ul)
)

Presence = Struct(
    "presence_detected" / Flag
)

EventStatistics = Struct(
    "number_of_events" / Count16,
    "average_event_duration" / TimeSecond16,
    "time_elapsed_since_last_event" / TimeExponential8,
    "sensing_duration" / TimeExponential8,
)

RelativeRuntimeInAGenericLevelRange = Struct(
    "relative_value" / DefaultCountValidator(Int8ul, rounding=1, resolution=0.5),
    "minimum_generic_level" / DefaultCountValidator(Int16ul),
    "maximum_generic_level" / DefaultCountValidator(Int16ul),
)

PropertyDict = {
    PropertyID.AVERAGE_AMBIENT_TEMPERATURE_IN_A_PERIOD_OF_DAY: Temperature8InAPeriodOfDay,
    PropertyID.AVERAGE_INPUT_CURRENT: AverageCurrent,
    PropertyID.AVERAGE_INPUT_VOLTAGE: AverageVoltage,
    PropertyID.AVERAGE_OUTPUT_CURRENT: AverageCurrent,
    PropertyID.AVERAGE_OUTPUT_VOLTAGE: AverageVoltage,
    PropertyID.CENTER_BEAM_INTENSITY_AT_FULL_POWER: LuminousIntensity,
    PropertyID.CHROMATICITY_TOLERANCE: ChromaticityTolerance,
    PropertyID.COLOR_RENDERING_INDEX_R9: ColorRenderingIndex,
    PropertyID.COLOR_RENDERING_INDEX_RA: ColorRenderingIndex,
    PropertyID.DEVICE_APPEARANCE: Appearance,
    PropertyID.DEVICE_COUNTRY_OF_ORIGIN: CountryCode,
    PropertyID.DEVICE_DATE_OF_MANUFACTURE: DateUTC,
    PropertyID.DEVICE_ENERGY_USE_SINCE_TURN_ON: Energy,
    PropertyID.DEVICE_FIRMWARE_REVISION: FixedString(8),
    PropertyID.DEVICE_GLOBAL_TRADE_ITEM_NUMBER: GlobalTradeItemNumber,
    PropertyID.DEVICE_HARDWARE_REVISION: FixedString(16),
    PropertyID.DEVICE_MANUFACTURER_NAME: FixedString(36),
    PropertyID.DEVICE_MODEL_NUMBER: FixedString(24),
    PropertyID.DEVICE_OPERATING_TEMPERATURE_RANGE_SPECIFICATION: TemperatureRange,
    PropertyID.DEVICE_OPERATING_TEMPERATURE_STATISTICAL_VALUES: TemperatureStatistics,
    PropertyID.DEVICE_OVER_TEMPERATURE_EVENT_STATISTICS: EventStatistics,
    PropertyID.DEVICE_POWER_RANGE_SPECIFICATION: PowerSpecification,
    PropertyID.DEVICE_RUNTIME_SINCE_TURN_ON: TimeHour24,
    PropertyID.DEVICE_RUNTIME_WARRANTY: TimeHour24,
    PropertyID.DEVICE_SERIAL_NUMBER: FixedString(16),
    PropertyID.DEVICE_SOFTWARE_REVISION: FixedString(8),
    PropertyID.DEVICE_UNDER_TEMPERATURE_EVENT_STATISTICS: EventStatistics,
    PropertyID.INDOOR_AMBIENT_TEMPERATURE_STATISTICAL_VALUES: Temperature8Statistics,
    PropertyID.INITIAL_CIE_1931_CHROMATICITY_COORDINATES: ChromaticityCoordinates,
    PropertyID.INITIAL_CORRELATED_COLOR_TEMPERATURE: CorrelatedColorTemperature,
    PropertyID.INITIAL_LUMINOUS_FLUX: LuminousFlux,
    PropertyID.INITIAL_PLANCKIAN_DISTANCE: ChromaticDistanceFromPlanckian,
    PropertyID.INPUT_CURRENT_RANGE_SPECIFICATION: ElectricCurrentSpecification,
    PropertyID.INPUT_CURRENT_STATISTICS: ElectricCurrentStatistics,
    PropertyID.INPUT_OVER_CURRENT_EVENT_STATISTICS: EventStatistics,
    PropertyID.INPUT_OVER_RIPPLE_VOLTAGE_EVENT_STATISTICS: EventStatistics,
    PropertyID.INPUT_OVER_VOLTAGE_EVENT_STATISTICS: EventStatistics,
    PropertyID.INPUT_UNDER_CURRENT_EVENT_STATISTICS: EventStatistics,
    PropertyID.INPUT_UNDER_VOLTAGE_EVENT_STATISTICS: EventStatistics,
    PropertyID.INPUT_VOLTAGE_RANGE_SPECIFICATION: VoltageRange,
    PropertyID.INPUT_VOLTAGE_RIPPLE_SPECIFICATION: Percentage8,
    PropertyID.INPUT_VOLTAGE_STATISTICS: VoltageStatistics,
    PropertyID.LIGHT_CONTROL_AMBIENT_LUXLEVEL_ON: Illuminance,
    PropertyID.LIGHT_CONTROL_AMBIENT_LUXLEVEL_PROLONG: Illuminance,
    PropertyID.LIGHT_CONTROL_AMBIENT_LUXLEVEL_STANDBY: Illuminance,
    PropertyID.LIGHT_CONTROL_LIGHTNESS_ON: PerceivedLightness,
    PropertyID.LIGHT_CONTROL_LIGHTNESS_PROLONG: PerceivedLightness,
    PropertyID.LIGHT_CONTROL_LIGHTNESS_STANDBY: PerceivedLightness,
    PropertyID.LIGHT_CONTROL_REGULATOR_ACCURACY: Percentage8,
    PropertyID.LIGHT_CONTROL_REGULATOR_KID: Coefficient,
    PropertyID.LIGHT_CONTROL_REGULATOR_KIU: Coefficient,
    PropertyID.LIGHT_CONTROL_REGULATOR_KPD: Coefficient,
    PropertyID.LIGHT_CONTROL_REGULATOR_KPU: Coefficient,
    PropertyID.LIGHT_CONTROL_TIME_FADE: TimeMiliseconds24,
    PropertyID.LIGHT_CONTROL_TIME_FADE_ON: TimeMiliseconds24,
    PropertyID.LIGHT_CONTROL_TIME_FADE_STANDBY_AUTO: TimeMiliseconds24,
    PropertyID.LIGHT_CONTROL_TIME_FADE_STANDBY_MANUAL: TimeMiliseconds24,
    PropertyID.LIGHT_CONTROL_TIME_OCCUPANCY_DELAY: TimeMiliseconds24,
    PropertyID.LIGHT_CONTROL_TIME_PROLONG: TimeMiliseconds24,
    PropertyID.LIGHT_CONTROL_TIME_RUN_ON: TimeMiliseconds24,
    PropertyID.LUMEN_MAINTENANCE_FACTOR: Percentage8,
    PropertyID.LUMINOUS_EFFICACY: LuminousEfficacy,
    PropertyID.LUMINOUS_ENERGY_SINCE_TURN_ON: LuminousEnergy,
    PropertyID.LUMINOUS_EXPOSURE: LuminousExposure,
    PropertyID.LUMINOUS_FLUX_RANGE: LuminousFluxRange,
    PropertyID.MOTION_SENSED: Percentage8,
    PropertyID.MOTION_THRESHOLD: Percentage8,
    PropertyID.OPEN_CIRCUIT_EVENT_STATISTICS: EventStatistics,
    PropertyID.OUTDOOR_STATISTICAL_VALUES: Temperature8Statistics,
    PropertyID.OUTPUT_CURRENT_RANGE: ElectricCurrentRange,
    PropertyID.OUTPUT_CURRENT_STATISTICS: ElectricCurrentStatistics,
    PropertyID.OUTPUT_RIPPLE_VOLTAGE_SPECIFICATION: Percentage8,
    PropertyID.OUTPUT_VOLTAGE_RANGE: VoltageRange,
    PropertyID.OUTPUT_VOLTAGE_STATISTICS: VoltageStatistics,
    PropertyID.OVER_OUTPUT_RIPPLE_VOLTAGE_EVENT_STATISTICS: EventStatistics,
    PropertyID.PEOPLE_COUNT: Count16,
    PropertyID.PRESENCE_DETECTED: Presence,
    PropertyID.PRESENT_AMBIENT_LIGHT_LEVEL: Illuminance,
    PropertyID.PRESENT_AMBIENT_TEMPERATURE: Temperature8,
    PropertyID.PRESENT_CIE_1931_CHROMATICITY_COORDINATES: ChromaticityCoordinates,
    PropertyID.PRESENT_CORRELATED_COLOR_TEMPERATURE: CorrelatedColorTemperature,
    PropertyID.PRESENT_DEVICE_INPUT_POWER: Power,
    PropertyID.PRESENT_DEVICE_OPERATING_EFFICIENCY: Percentage8,
    PropertyID.PRESENT_DEVICE_OPERATING_TEMPERATURE: Temperature,
    PropertyID.PRESENT_ILLUMINANCE: Illuminance,
    PropertyID.PRESENT_INDOOR_AMBIENT_TEMPERATURE: Temperature8,
    PropertyID.PRESENT_INPUT_CURRENT: ElectricCurrent,
    PropertyID.PRESENT_INPUT_RIPPLE_VOLTAGE: Percentage8,
    PropertyID.PRESENT_INPUT_VOLTAGE: Voltage,
    PropertyID.PRESENT_LUMINOUS_FLUX: LuminousFlux,
    PropertyID.PRESENT_OUTDOOR_AMBIENT_TEMPERATURE: Temperature8,
    PropertyID.PRESENT_OUTPUT_CURRENT: ElectricCurrent,
    PropertyID.PRESENT_OUTPUT_VOLTAGE: Voltage,
    PropertyID.PRESENT_PLANCKIAN_DISTANCE: ChromaticDistanceFromPlanckian,
    PropertyID.PRESENT_RELATIVE_OUTPUT_RIPPLE_VOLTAGE: Percentage8,
    PropertyID.RELATIVE_DEVICE_ENERGY_USE_IN_A_PERIOD_OF_DAY: EnergyInAPeriodOfDay,
    PropertyID.RELATIVE_DEVICE_RUNTIME_IN_A_GENERIC_LEVEL_RANGE: RelativeRuntimeInAGenericLevelRange,
    PropertyID.RELATIVE_EXPOSURE_TIME_IN_AN_ILLUMINANCE_RANGE: RelativeValueInAnIlluminanceRange,
    PropertyID.RELATIVE_RUNTIME_IN_A_CORRELATED_COLOR_TEMPERATURE_RANGE: LuminousEnergy,
    PropertyID.RELATIVE_RUNTIME_IN_A_DEVICE_OPERATING_TEMPERATURE_RANGE: RelativeValueInATemperatureRange,
    PropertyID.RELATIVE_RUNTIME_IN_AN_INPUT_CURRENT_RANGE: RelativeValueInACurrentRange,
    PropertyID.RELATIVE_RUNTIME_IN_AN_INPUT_VOLTAGE_RANGE: RelativeValueInAVoltageRange,
    PropertyID.SHORT_CIRCUIT_EVENT_STATISTICS: EventStatistics,
    PropertyID.TIME_SINCE_MOTION_SENSED: TimeSecond16,
    PropertyID.TIME_SINCE_PRESENCE_DETECTED: TimeSecond16,
    PropertyID.TOTAL_DEVICE_ENERGY_USE: Energy,
    PropertyID.TOTAL_DEVICE_OFF_ON_CYCLES: Count24,
    PropertyID.TOTAL_DEVICE_POWER_ON_CYCLES: Count24,
    PropertyID.TOTAL_DEVICE_POWER_ON_TIME: TimeHour24,
    PropertyID.TOTAL_DEVICE_RUNTIME: TimeHour24,
    PropertyID.TOTAL_LIGHT_EXPOSURE_TIME: TimeHour24,
    PropertyID.TOTAL_LUMINOUS_ENERGY: LuminousEnergy,
    PropertyID.PRECISE_TOTAL_DEVICE_ENERGY_USE: PreciseEnergy,
}

PropertyValue = Switch(
    this.sensor_setting_property_id,
        PropertyDict,
        default=Array(this.length, Byte)
)
# fmt: off



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest
from construct import ValidationError

from bluetooth_mesh.messages.generic.onoff import *
from bluetooth_mesh.messages.generics import *


@pytest.mark.parametrize(
    # fmt: off
    "steps, resolution, expected_tt",
    [
        (0x00, 0b00, "00"),
        (0x01, 0b00, "01"),
        (0x01, 0b01, "41"),
        (0x3E, 0b00, "3E"),
        (0x3E, 0b11, "FE"),
        (0x3F, 0b11, "FF")
    ]
    # fmt: on
)
def test_build_transition_time(steps, resolution, expected_tt):
    tt = TransitionTime.build(dict(steps=steps, resolution=resolution))
    assert tt == bytes.fromhex(expected_tt)


@pytest.mark.parametrize(
    # fmt: off
    "raw_bytes, expected_steps, expected_resolution",
    [
        ("00", 0x00, 0b00),
        ("01", 0x01, 0b00),
        ("40", 0x00, 0b01),
        ("3E", 0x3E, 0b00),
        ("FE", 0x3E, 0b11),
        ("FF", 0x3F, 0b11)
    ]
    # fmt: on
)
def test_parse_transition_time(raw_bytes, expected_steps, expected_resolution):
    tt = TransitionTime.parse(bytes.fromhex(raw_bytes))
    assert tt.steps == expected_steps
    assert tt.resolution == expected_resolution


valid = [
    # fmt: off
    pytest.param(
        b'\x82\x01',
        GenericOnOffOpcode.ONOFF_GET,
        dict(),
        id="ONOFF_GET"),
    pytest.param(
        b'\x82\x02\x01\x22',
        GenericOnOffOpcode.ONOFF_SET,
        dict(
            onoff=1,
            tid=34
        ),
        id="ONOFF_SET"),
    pytest.param(
        b'\x82\x02\x01\x22',
        GenericOnOffOpcode.ONOFF_SET,
        dict(
            onoff=1,
            tid=34,
            transition_time=6.3,
            delay=0.3
        ),
        id="ONOFF_SET_invalid"),
    pytest.param(
        b'\x82\x02\x00\x31\x32\x3c',
        GenericOnOffOpcode.ONOFF_SET,
        dict(
            onoff=0,
            tid=49,
            transition_time=5,
            delay=0.3
        ),
        id="ONOFF_SET_with_optional"),
    pytest.param(
        b'\x82\x02\x00\x31\xf2\x3c',
        GenericOnOffOpcode.ONOFF_SET,
        dict(
            onoff=0,
            tid=49,
            transition_time=30000,
            delay=0.3
        ),
        id="ONOFF_SET_large_transition_time"),
    pytest.param(
        b'\x82\x04\x00',
        GenericOnOffOpcode.ONOFF_STATUS,
        dict(
            present_onoff=0,
            target_onoff=None,
            remaining_time=None
        ),
        id="ONOFF_STATUS"
    ),
    pytest.param(
        b'\x82\x04\x00\x01\x4a',
        GenericOnOffOpcode.ONOFF_STATUS,
        dict(
            present_onoff=0,
            target_onoff=1,
            remaining_time=10
        ),
        id="ONOFF_STATUS_with_optional"
    ),
    pytest.param(
        b'\x82\x04\x00\x01\xff',
        GenericOnOffOpcode.ONOFF_STATUS,
        dict(
            present_onoff=0,
            target_onoff=1,
            remaining_time=37800
        ),
        id="ONOFF_STATUS_max_time"
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_parse_valid(encoded, opcode, data):
    assert GenericOnOffMessage.parse(encoded).params == data


@pytest.mark.parametrize("encoded,opcode,data", valid)
def test_build_valid(encoded, opcode, data):
    assert GenericOnOffMessage.build(dict(opcode=opcode, params=data)) == encoded



from functools import partial

from bluetooth_mesh.messages.health import HealthOpcode


async def attention_unack(self, destination, app_index, attention):
    # create a argument-less callable that sends an application message
    request = partial(
        self.send_app,
        destination,
        app_index=app_index,
        opcode=HealthOpcode.ATTENTION_SET_UNACKNOWLEDGED,
        params=dict(
            attention=attention,
        )
    )

    # send 3 *repeated* messages with 0.5s interval
    await self.repeat(request, retransmissions=3, send_interval=0.5)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import enum

from construct import (
    Default,
    ExprValidator,
    GreedyRange,
    Int8ul,
    Int16ul,
    Struct,
    Switch,
    obj_,
    this,
)

from .util import EnumAdapter, Opcode

# fmt: off
FaultTest = Struct(
    "test_id" / Int8ul,
    "company_id" / Int16ul,
)

FaultStatus = Struct(
    "test_id" / Int8ul,
    "company_id" / Int16ul,
    "fault_array" / Default(GreedyRange(Int8ul), [])
)

FastPeriodDivisor = Struct(
    "fast_period_divisor" / ExprValidator(Int8ul, obj_ <= 15)
)

Attention = Struct(
    "attention" / Int8ul
)

CompanyId = Struct(
    "company_id" / Int16ul
)

HealthCurrentStatus = FaultStatus

HealthFaultGet = CompanyId

HealthFaultClear = CompanyId

HealthFaultTest = FaultTest

HealthFaultStatus = FaultStatus

HealthPeriodGet = Struct()

HealthPeriodSet = FastPeriodDivisor

HealthPeriodStatus = FastPeriodDivisor

HealthAttentionGet = Struct()

HealthAttentionSet = Attention

HealthAttentionStatus = Attention
# fmt: off

class HealthOpcode(enum.IntEnum):
    ATTENTION_GET = 0x8004
    ATTENTION_SET = 0x8005
    ATTENTION_SET_UNACKNOWLEDGED = 0x8006
    ATTENTION_STATUS = 0x8007
    CURRENT_STATUS = 0x04
    FAULT_CLEAR = 0x802F
    FAULT_CLEAR_UNACKNOWLEDGED = 0x8030
    FAULT_GET = 0x8031
    FAULT_STATUS = 0x05
    FAULT_TEST = 0x8032
    FAULT_TEST_UNACKNOWLEDGED = 0x8033
    PERIOD_GET = 0x8034
    PERIOD_SET = 0x8035
    PERIOD_SET_UNACKNOWLEDGED = 0x8036
    PERIOD_STATUS = 0x8037


# fmt: off
HealthMessage = Struct(
    "opcode" / Opcode(HealthOpcode),
    "params" / Switch(
        this.opcode,
        {
            HealthOpcode.ATTENTION_GET: HealthAttentionGet,
            HealthOpcode.ATTENTION_SET: HealthAttentionSet,
            HealthOpcode.ATTENTION_SET_UNACKNOWLEDGED: HealthAttentionSet,
            HealthOpcode.ATTENTION_STATUS: HealthAttentionStatus,
            HealthOpcode.CURRENT_STATUS: HealthCurrentStatus,
            HealthOpcode.FAULT_CLEAR: HealthFaultClear,
            HealthOpcode.FAULT_CLEAR_UNACKNOWLEDGED: HealthFaultClear,
            HealthOpcode.FAULT_GET: HealthFaultGet,
            HealthOpcode.FAULT_STATUS: HealthCurrentStatus,
            HealthOpcode.FAULT_TEST: HealthFaultTest,
            HealthOpcode.FAULT_TEST_UNACKNOWLEDGED: HealthFaultTest,
            HealthOpcode.PERIOD_GET: HealthPeriodGet,
            HealthOpcode.PERIOD_SET: HealthPeriodSet,
            HealthOpcode.PERIOD_SET_UNACKNOWLEDGED: HealthPeriodSet,
            HealthOpcode.PERIOD_STATUS: HealthPeriodStatus,
        }
    )
)
# fmt: on



from typing import Any, Callable, List, Tuple

from construct import (
    Adapter,
    Array,
    BitsInteger,
    Bytes,
    BytesInteger,
    Computed,
    Const,
    Default,
    Embedded,
    Enum,
    FixedSized,
    Flag,
    FormatField,
    GreedyBytes,
    GreedyRange,
    NullStripped,
    Padded,
    Pass,
    Rebuild,
    Renamed,
    Restreamed,
    Select,
    StringEncoded,
    Struct,
    Switch,
    Transformed,
)
from docutils import nodes
from sphinx.ext.autodoc import DataDocumenter, members_option
from sphinx.util import logging

logger = logging.getLogger(__name__)


def level_option(arg):
    return int(arg)


class Skip:
    pass


class ConstructDocumenter(DataDocumenter):
    """
    Specialized Documenter subclass for construct items.
    """
    objtype = 'construct'
    directivetype = 'data'

    option_spec = dict(DataDocumenter.option_spec)
    option_spec["level"] = level_option

    INDENT = ' ' * 4

    def document_struct(self, subcon, level):
        doc = []

        for field in subcon.subcons:
            annotation, subdoc = self.document_subcon(field, level + 1)

            if annotation is not Skip:
                annotation = ': %s' % annotation if annotation else ''
                name = '`%s`' % field.name if field.name else ''
                if name and annotation:
                    doc.append('%s* %s%s' % (self.INDENT * level, name, annotation))

            doc += subdoc

        return ':py:class:`Struct` of', doc

    def document_switch(self, subcon, level):
        doc = []
        for name, field in subcon.cases.items():
            _, subdoc = self.document_subcon(field, level + 1)

            doc.append('%s* :py:data:`%s`' % (self.INDENT * level, name))
            doc += subdoc

        return ':py:class:`Switch` by `%s`' % subcon.keyfunc.__getfield__(), doc

    def document_enum(self, subcon, level):
        annotation, subdoc = self.document_subcon(subcon.subcon, level + 1)
        return ':py:class:`Enum` of :py:class:`%s` as %s' % (subcon.ENUM.__name__, annotation), subdoc

    def document_const(self, subcon):
        return ':py:class:`Const` :py:data:`%s`' % subcon.value, []

    def document_select(self, subcon, level):
        doc = []

        for field in subcon.subcons:
            annotation, subdoc = self.document_subcon(field, level + 1)

            if annotation is not Skip:
                annotation = ': %s' % annotation if field.name and annotation else annotation
                name = ':py:class:`%s`' % field.name if field.name else ''
                if name or annotation:
                    doc.append('%s* %s%s' % (self.INDENT * level, name, annotation))

            doc += subdoc

        return ':py:class:`Select` from', doc

    def document_array(self, subcon, level):
        annotation, subdoc = self.document_subcon(subcon.subcon, level + 1)
        if isinstance(subcon.count, Callable):
            return ':py:class:`Array[%s]` of %s' % (subcon.count.__getfield__(), annotation), subdoc
        else:
            return ':py:class:`Array[%i]` of %s' % (subcon.count, annotation), subdoc

    def document_format_field(self, subcon):
        DOC = {
            "B": "Uint8",
            "H": "Uint16",
            "L": "Uint32",
            "Q": "Uint64",
            "b": "Int8",
            "h": "Int16",
            "l": "Int32",
            "q": "Int64",
        }

        return ':py:class:`%s`' % DOC[subcon.fmtstr[1]], []

    def document_bytes_integer(self, subcon):
        return ':py:class:`{}nt{}`'.format('I' if subcon.signed else 'Ui', subcon.length * 8), []

    def document_bits_integer(self, subcon):
        return ':py:class:`{}nt{}`'.format('I' if subcon.signed else 'Ui', subcon.length), []

    def document_bytes(self, subcon):
        return ':py:class:`bytes[%i]`' % subcon.length, []

    def document_range(self, subcon, level):
        annotation, subdoc = self.document_subcon(subcon.subcon, level + 1)
        return ':py:class:`Range` of %s' % annotation, subdoc

    def document_subcon(self, subcon, level=0):
        if self.options.level is not None and level > self.options.level:
            return '', []

        if hasattr(subcon, '__construct_doc__'):
            return self.document_subcon(subcon.__construct_doc__, level)

        if isinstance(subcon, Enum):
            return self.document_enum(subcon, level)

        if isinstance(subcon, (Renamed, Adapter, Default, Transformed,
                               Restreamed, Rebuild, FixedSized,
                               NullStripped)):
            return self.document_subcon(subcon.subcon, level)

        if isinstance(subcon, Embedded):
            annotation, subdoc = self.document_subcon(subcon.subcon, level - 1)
            return Skip, subdoc

        if isinstance(subcon, Struct):
            return self.document_struct(subcon, level)

        if isinstance(subcon, Switch):
            return self.document_switch(subcon, level)

        if isinstance(subcon, Select):
            return self.document_select(subcon, level)

        if isinstance(subcon, Array):
            return self.document_array(subcon, level)

        if isinstance(subcon, GreedyRange):
            return self.document_range(subcon, level)

        if isinstance(subcon, FormatField):
            return self.document_format_field(subcon)

        if isinstance(subcon, BytesInteger):
            return self.document_bytes_integer(subcon)

        if isinstance(subcon, BitsInteger):
            return self.document_bits_integer(subcon)

        if isinstance(subcon, Bytes):
            return self.document_bytes(subcon)

        if isinstance(subcon, Padded) and not subcon.name:
            return Skip, []

        if isinstance(subcon, Computed):
            return Skip, []

        if isinstance(subcon, Const):
            return self.document_const(subcon)

        if isinstance(subcon, StringEncoded):
            return ':py:class:`str`', []

        if subcon is Pass:
            return ':py:data:`None`', []

        if subcon is Flag:
            return ':py:class:`bool`', []

        if subcon is GreedyBytes:
            return ':py:class:`bytes`', []

        logger.warn('Uknown subconstruct: %s %s', subcon.name, subcon)

        return ': :py:class:`UNKNOWN`', []

    def get_doc(self, encoding: str = None, ignore: int = 1) -> List[List[str]]:
        _, subdoc = self.document_subcon(self.object)
        return [subdoc]


def setup(app):
    app.add_autodocumenter(ConstructDocumenter)

    return {'version': '0.1', 'parallel_read_safe': True}



from enum import IntEnum

from bluetooth_mesh.messages.generic.onoff import GenericOnOffOpcode
from bluetooth_mesh.models import Model


class GenericOnOffServer(Model):
    MODEL_ID = (None, 0x1002)
    OPCODES = [
        GenericOnOffOpcode.ONOFF_SET,
        GenericOnOffOpcode.ONOFF_SET_UNACKNOWLEDGED,
        GenericOnOffOpcode.ONOFF_GET,
    ]


MY_VENDOR_ID = 0x0136


class MyVendorModelOpcode(IntEnum):
    FROBNICATOR_GET = 0x0001


class MyVendorModel(Model):
    MODEL_ID = (MY_VENDOR_ID, 0x0001)
    OPCODES = [
        MyVendorModelOpcode.FROBNICATOR_GET,
    ]



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.silvair.debug import DebugPayload, DebugSubOpcode

# fmt: off
valid = [
    pytest.param(b'\x00', DebugSubOpcode.RSSI_THRESHOLD_GET, None, id="RssiThresholdGet"),
    pytest.param(b'\x04', DebugSubOpcode.TIMESLOT_TX_POWER_GET, None, id="TimeslotTxPowerGet"),
    pytest.param(b'\x07', DebugSubOpcode.SOFTDEVICE_TX_POWER_GET, None, id="SoftDeviceTxPowerGet"),
    pytest.param(b'\x0a', DebugSubOpcode.UPTIME_GET, None, id="UptimeGet"),
    pytest.param(b'\x0c', DebugSubOpcode.LAST_SW_FAULT_GET, None, id="LastSwFaultGet"),
    pytest.param(b'\x0f', DebugSubOpcode.SYSTEM_STATS_GET, None, id="SystemStatsGet"),
    pytest.param(b'\x11', DebugSubOpcode.LAST_MALLOC_FAULT_GET, None, id="LastMallocFaultGet"),
    pytest.param(b'\x14', DebugSubOpcode.LAST_FDS_FAULT_GET, None, id="LastFdsFaultGet"),
    pytest.param(b'\x17', DebugSubOpcode.BYTES_BEFORE_GARBAGE_COLLECTOR_GET, None, id="BytesBeforeGCGet"),
    pytest.param(b'\x19', DebugSubOpcode.PROVISIONED_APP_VERSION_GET, None, id="ProvisionedAppVersionGet"),
    pytest.param(b'\x1b', DebugSubOpcode.FULL_FIRMWARE_VERSION_GET, None, id="FullFirmwareVersionGet"),
    pytest.param(b'\x1d', DebugSubOpcode.IV_INDEX_GET, None, id="IvIndexGet"),
    pytest.param(b'\x1f', DebugSubOpcode.GARBAGE_COLLECTOR_COUNTER_GET, None, id="GCCounterGet"),
    pytest.param(b'\x0d', DebugSubOpcode.LAST_SW_FAULT_CLEAR, None, id="LastSwFaultClear"),
    pytest.param(b'\x12', DebugSubOpcode.LAST_MALLOC_FAULT_CLEAR, None, id="LastMallocFaultClear"),
    pytest.param(b'\x15', DebugSubOpcode.LAST_FDS_FAULT_CLEAR, None, id="FastFdsFaultClear"),
    pytest.param(b'\x21', DebugSubOpcode.ARAP_LIST_SIZE_GET, None, id="ARAPListSizeGet"),
    pytest.param(
        b'\x23\x02',
        DebugSubOpcode.ARAP_LIST_CONTENT_GET,
        {'page': 2},
        id="ARAPListContentGet"
    ),
    pytest.param(
        b'\x01\x80',
        DebugSubOpcode.RSSI_THRESHOLD_SET,
        {'rssi_threshold': 0x80},
        id="RssiThresholdGet"
    ),
    pytest.param(
        b'\x02\x80',
        DebugSubOpcode.RSSI_THRESHOLD_STATUS,
        {'rssi_threshold': 0x80},
        id="RssiThresholdStatus"
    ),
    pytest.param(
        b'\x03\x01',
        DebugSubOpcode.RADIO_TEST,
        {'packet_counter': 0x01},
        id="RadioTest"
    ),
    pytest.param(
        b'\x05\x02',
        DebugSubOpcode.TIMESLOT_TX_POWER_SET,
        {'tx_power': 0x02},
        id="TimeslotTxPowerSet"
    ),
    pytest.param(
        b'\x06\x04',
        DebugSubOpcode.TIMESLOT_TX_POWER_STATUS,
        {'tx_power': 0x04},
        id="TimeslotTxPowerStatus"
    ),
    pytest.param(
        b'\x08\x02',
        DebugSubOpcode.SOFTDEVICE_TX_POWER_SET,
        {'tx_power': 0x02},
        id="SoftDeviceTxPowerSet"
    ),
    pytest.param(
        b'\x09\x04',
        DebugSubOpcode.SOFTDEVICE_TX_POWER_STATUS,
        {'tx_power': 0x04},
        id="SoftDeviceTxPowerStatus"
    ),
    pytest.param(
        b'\x0b\xa2\xd0\x02\x00',
        DebugSubOpcode.UPTIME_STATUS,
        {'uptime': 184482},
        id="UptimeStatus"
    ),
    pytest.param(
        b'\x0e\x0a\x00\x00\x00Power OFF [7]',
        DebugSubOpcode.LAST_SW_FAULT_STATUS,
        {'time': 10, 'fault': 'Power OFF [7]'},
        id="LastSwFaultStatus"
    ),
    pytest.param(
        b'\x13\x00\x00\x00\x00No Error',
        DebugSubOpcode.LAST_MALLOC_FAULT_STATUS,
        {'time': 0, 'fault': 'No Error'},
        id="LastMallocFaultStatus"
    ),
    pytest.param(
        b'\x16\xff\x00\x00\x00Foo error',
        DebugSubOpcode.LAST_FDS_FAULT_STATUS,
        {'time': 255, 'fault': 'Foo error'},
        id="LastFdsFaultStatus"
    ),
    pytest.param(
        b'\x18\xf8.',
        DebugSubOpcode.BYTES_BEFORE_GARBAGE_COLLECTOR_STATUS,
        {'bytes_left': 12024},
        id="BytesBeforeGarbageCollectorStatus"
    ),
    pytest.param(
        b'\x1a\x08\x0c',
        DebugSubOpcode.PROVISIONED_APP_VERSION_STATUS,
        {'version': 3080},
        id="ProvisionedAppVersionStatus"
    ),
    pytest.param(
        b'\x1c2.12.0-rc4-6c74b4d',
        DebugSubOpcode.FULL_FIRMWARE_VERSION_STATUS,
        {'version': '2.12.0-rc4-6c74b4d'},
        id="FullFirmwareVersionStatus"
    ),
    pytest.param(
        b'\x1e\x05\x00\x00\x00',
        DebugSubOpcode.IV_INDEX_STATUS,
        {'ivindex': 5},
        id="IvIndexStatus"
    ),
    pytest.param(
        b' \x01\x00',
        DebugSubOpcode.GARBAGE_COLLECTOR_COUNTER_STATUS,
        {'counter': 1},
        id="GarbageCollectorCounterStatus"
    ),
    pytest.param(
        bytes.fromhex("22 FEFF 0100"),
        DebugSubOpcode.ARAP_LIST_SIZE_STATUS,
        {'capacity': 65534, 'size': 1},
        id="ArapListSizeStatus"
    ),
    pytest.param(
        b'\x10'
        b'KERNEL\x00\x00\x08\x02\x00\x00\x00\x00'
        b'BLE\x00\x00\x00\x00\x00\x08\x05\x00\x00\x00\x00'
        b'SNV\x00\x00\x00\x00\x00\xc8\x00\x00\x00\x00\x00'
        b'APP\x00\x00\x00\x00\x00\xb4\x01\x00\x00\x00\x00'
        b'RX Task\x00\xe4\x01\x00\x00\x00\x00'
        b'MESH\x00\x00\x00\x00\x08\x04\x00\x00\x00\x00'
        b'Tmr Svc\x00\xf0\x01\x00\x00\x00\x00'
        b'IDLE\x00\x00\x00\x00\xe4\x00\x00\x00\x00\x00'
        b'HEAP\x00\x00\x00\x00\xa8I\x00\x00\x00\x00',
        DebugSubOpcode.SYSTEM_STATS_STATUS,
        {
            "stats": {
                'KERNEL': 520,
                'BLE': 1288,
                'SNV': 200,
                'APP': 436,
                'RX Task': 484,
                'MESH': 1032,
                'Tmr Svc': 496,
                'IDLE': 228,
                'HEAP': 18856,
            }
        },
        id='SystemStatsStatus (long)'
    ),
    pytest.param(
        b'\x10'
        b'HEAP\x00\x00\x00\x00\x88J\x00\x00\x00\x00'
        b'Tmr Svc\x00(\x02\x00\x00\x00\x00'
        b'IDLE\x00\x00\x00\x00\xf4\x00\x00\x00\x00\x00',
        DebugSubOpcode.SYSTEM_STATS_STATUS,
        {
            "stats": {
                'HEAP': 19080, 'Tmr Svc': 552, 'IDLE': 244,
            }
        },
        id='SystemStatsStatus (short)'
    ),
    pytest.param(
        b'$\x00\x00\x80\x80X\x00\x00',
        DebugSubOpcode.ARAP_LIST_CONTENT_STATUS,
        {
            'current_page': 0,
            'last_page': 0,
            'nodes': {
                0x0080: {"ivi": True, "sequence": 0x58},
            }
        },
        id='ArapContent (1 node)'
    ),
    pytest.param(
        b'$\x00\x00\x07\x00p\x0c\x02\x80\x00X\x00\x00',
        DebugSubOpcode.ARAP_LIST_CONTENT_STATUS,
        {
            'current_page': 0,
            'last_page': 0,
            'nodes': {
                0x0007: {"ivi": False, "sequence": 0x20c70},
                0x0080: {"ivi": False, "sequence": 0x58},
            }
        },
        id='ArapContent (2 nodes)'
    ),
]

valid_parse = [
    pytest.param(
        bytes.fromhex("22 7E 0E"),
        DebugSubOpcode.ARAP_LIST_SIZE_STATUS,
        {'capacity': 126, 'size': 14},
        id="ArapListSizeStatus")
]

invalid_build = [
    pytest.param(
        bytes.fromhex("22 7E 0E"),
        DebugSubOpcode.ARAP_LIST_SIZE_STATUS,
        {'capacity': 126, 'size': 14},
        id="ArapListSizeStatus",
        marks=pytest.mark.xfail(reason="1 byte ARAP List Size is not supported in latest FW", strict=True))
]
# fmt: on


@pytest.mark.parametrize("encoded,subopcode,data", valid + valid_parse)
def test_parse_valid(encoded, subopcode, data):
    assert DebugPayload.parse(encoded) == dict(subopcode=subopcode, data=data)


@pytest.mark.parametrize("encoded,subopcode,data", valid + invalid_build)
def test_build_valid(encoded, subopcode, data):
    assert DebugPayload.build(dict(subopcode=subopcode, data=data)) == encoded



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.messages.silvair.gateway_config_server import (
    ConnState,
    DhcpFlag,
    GatewayConfigPayload,
    GatewayConfigServerSubOpcode,
    LastError,
    LinkStatus,
    StatusCode,
)
from bluetooth_mesh.messages.util import IPv4Address

valid = [
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("00"),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_GET,
        None,
        id="ConfigurationGet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("01 B004 ABCDEFABCDEF D204 D007 0B 3139322E3136382E302E31"),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
        dict(
            mtu_size=1200,
            mac_address="ab:cd:ef:ab:cd:ef",
            server_port_number=1234,
            reconnect_interval=2000,
            server_address_length=11,
            server_address="192.168.0.1",
        ),
        id="ConfigurationSetWithoutOptionalAutoDhcpEnabled",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "01 B004 ABCDEFABCDEF D204 D007 19 7777772E72616E646F6D686F73746E616D652E636F6D2E706C"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
        dict(
            mtu_size=1200,
            mac_address="ab:cd:ef:ab:cd:ef",
            server_port_number=1234,
            reconnect_interval=2000,
            server_address_length=25,
            server_address="www.randomhostname.com.pl",
        ),
        id="ConfigurationSetWithoutOptionalAutoDhcpEnabledHostname",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "01 B004 ABCDEFABCDEF D204 D007 0B 3139322E3136382E302E31 0A2A0002"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
        dict(
            mtu_size=1200,
            mac_address="ab:cd:ef:ab:cd:ef",
            server_port_number=1234,
            reconnect_interval=2000,
            server_address_length=11,
            server_address="192.168.0.1",
            dns_ip_address=IPv4Address("10.42.0.2"),
        ),
        id="ConfigurationSetWithOptionalDhcpEnabledWithStaticDns",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "01 DE03 ABCDEFABCDEF ABCD 04D2"
            "19 7777772E72616E646F6D686F73746E616D652E636F6D2E706C 0A2A0002"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
        dict(
            mtu_size=990,
            mac_address="ab:cd:ef:ab:cd:ef",
            server_port_number=52651,
            reconnect_interval=53764,
            server_address_length=25,
            server_address="www.randomhostname.com.pl",
            dns_ip_address=IPv4Address("10.42.0.2"),
        ),
        id="ConfigurationSetWithOptionalDhcpEnabledWithStaticDnsHostname",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "01 B004 ABCDEFABCDEF D204 D007 0B 3139322E3136382E302E31 0A2A0002 C0A80002 C0A80001 08"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
        dict(
            mtu_size=1200,
            mac_address="ab:cd:ef:ab:cd:ef",
            server_port_number=1234,
            reconnect_interval=2000,
            server_address_length=11,
            server_address="192.168.0.1",
            dns_ip_address=IPv4Address("10.42.0.2"),
            ip_address=IPv4Address("192.168.0.2"),
            gateway_ip_address=IPv4Address("192.168.0.1"),
            netmask=8,
        ),
        id="ConfigurationSetWithOptionalDhcpDisabled",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "01 B004 ABCDEFABCDEF D204 D007 19 7777772E72616E646F6D686F73746E616D652E636F6D2E706C"
            "0A2A0002 C0A80002 C0A80001 08"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_SET,
        dict(
            mtu_size=1200,
            mac_address="ab:cd:ef:ab:cd:ef",
            server_port_number=1234,
            reconnect_interval=2000,
            server_address_length=25,
            server_address="www.randomhostname.com.pl",
            dns_ip_address=IPv4Address("10.42.0.2"),
            ip_address=IPv4Address("192.168.0.2"),
            gateway_ip_address=IPv4Address("192.168.0.1"),
            netmask=8,
        ),
        id="ConfigurationSetWithOptionalDhcpDisabledHostname",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("02"),
        GatewayConfigServerSubOpcode.GATEWAY_PACKETS_GET,
        None,
        id="PacketsGet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("03"),
        GatewayConfigServerSubOpcode.GATEWAY_PACKETS_CLEAR,
        None,
        id="PacketsClear",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("04 DC05"),
        GatewayConfigServerSubOpcode.MTU_SIZE_SET,
        dict(mtu_size=1500),
        id="mtu_size",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("05 112233445566"),
        GatewayConfigServerSubOpcode.ETHERNET_MAC_ADDRESS_SET,
        dict(mac_address="11:22:33:44:55:66"),
        id="EthernetMacAddressSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("06 4522 0B 3139322E3136382E302E31"),
        GatewayConfigServerSubOpcode.SERVER_ADDRESS_AND_PORT_NUMBER_SET,
        dict(
            server_port_number=8773,
            server_address_length=11,
            server_address="192.168.0.1",
        ),
        id="ServerAddressAndPortNumberSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("06 4522 19 7777772E72616E646F6D686F73746E616D652E636F6D2E706C"),
        GatewayConfigServerSubOpcode.SERVER_ADDRESS_AND_PORT_NUMBER_SET,
        dict(
            server_port_number=8773,
            server_address_length=25,
            server_address="www.randomhostname.com.pl",
        ),
        id="ServerAddressAndPortNumberSetHostname",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("07 E803"),
        GatewayConfigServerSubOpcode.RECONNECT_INTERVAL_SET,
        dict(reconnect_interval=1000),
        id="ReconnectIntervalSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("08 0A2A0002"),
        GatewayConfigServerSubOpcode.DNS_IP_ADDRESS_SET,
        dict(dns_ip_address=IPv4Address("10.42.0.2")),
        id="DnsIpAddressSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("09 C0A80001"),
        GatewayConfigServerSubOpcode.IP_ADDRESS_SET,
        dict(ip_address=IPv4Address("192.168.0.1")),
        id="IpAddressSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("0A C0A80C03"),
        GatewayConfigServerSubOpcode.GATEWAY_IP_ADDRESS_SET,
        dict(gateway_ip_address=IPv4Address("192.168.12.03")),
        id="GatewayIpAddressSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("0B 08"),
        GatewayConfigServerSubOpcode.NETMASK_SET,
        dict(netmask=8),
        id="NetmaskSet",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "0C 0A B202 DDDDEFABCDEF 1111 2222 0F 7777772E74657374696E672E636F6D 7B2C0515 C0A80A02 C0A80A01 0A 00 00"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
        dict(
            chip_revision_id=10,
            mtu_size=690,
            mac_address="dd:dd:ef:ab:cd:ef",
            server_port_number=4369,
            reconnect_interval=8738,
            server_address_length=15,
            server_address="www.testing.com",
            dns_ip_address=IPv4Address("123.44.5.21"),
            ip_address=IPv4Address("192.168.10.2"),
            gateway_ip_address=IPv4Address("192.168.10.1"),
            netmask=10,
            flags=DhcpFlag.DHCP_DISABLED,
            status_code=StatusCode.STATUS_SUCCESS,
        ),
        id="ConfigurationStatus",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex(
            "0C 0A B202 DDDDEFABCDEF 1111 2222 00 C0A80A02 C0A80A02 C0A80A01 0A 02 00"
        ),
        GatewayConfigServerSubOpcode.GATEWAY_CONFIGURATION_STATUS,
        dict(
            chip_revision_id=10,
            mtu_size=690,
            mac_address="dd:dd:ef:ab:cd:ef",
            server_port_number=4369,
            reconnect_interval=8738,
            server_address_length=0,
            server_address="",
            dns_ip_address=IPv4Address("192.168.10.2"),
            ip_address=IPv4Address("192.168.10.2"),
            gateway_ip_address=IPv4Address("192.168.10.01"),
            netmask=10,
            flags=DhcpFlag.DHCP_ENABLED_AUTO_DNS,
            status_code=StatusCode.STATUS_SUCCESS,
        ),
        id="ConfigurationStatusEmptyHostname",
    ),
    pytest.param(
        GatewayConfigPayload,
        bytes.fromhex("0D 0000 0002 03E8 92"),
        GatewayConfigServerSubOpcode.GATEWAY_PACKETS_STATUS,
        dict(
            total_eth_rx_errors=0,
            total_eth_tx_errors=512,
            bandwidth=59395,
            connection_state=dict(
                conn_state=ConnState.ETHERNET_HANDSHAKE,
                link_status=LinkStatus.LINK_UP,
                last_error=LastError.ERROR_ENODEV,
            ),
        ),
        id="PacketsStatus",
    ),
]


@pytest.mark.parametrize("test, encoded, subopcode, payload", valid)
def test_parse_valid(test, encoded, subopcode, payload):
    print(test.parse(encoded))
    print(dict(subopcode=subopcode, payload=payload))
    assert test.parse(encoded) == dict(subopcode=subopcode, payload=payload)



from functools import partial

from bluetooth_mesh.messages.debug import DebugOpcode, DebugSubOpcode


async def get_uptime(self, nodes, net_index):
    # dictionary of request callables
    requests = {
        node: partial(
            self.send_dev,
            node,
            net_index=net_index,
            opcode=DebugOpcode.SILVAIR_DEBUG,
            params=dict(subopcode=DebugSubOpcode.UPTIME_GET)
        ) for node in nodes
    }

    # dictionary of Futures
    statuses = {
        node: self.expect_dev(
            node,
            net_index=0,
            opcode=DebugOpcode.SILVAIR_DEBUG,
            params=dict(subopcode=DebugSubOpcode.UPTIME_STATUS)
        ) for node in nodes
    }

    return await self.bulk_query(requests,
                                 statuses,
                                 send_interval=0.1,
                                 timeout=5.0)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import asyncio
import inspect
from collections import defaultdict
from contextlib import suppress
from datetime import timedelta
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Hashable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)
from uuid import UUID

from bluetooth_mesh.messages import AccessMessage
from bluetooth_mesh.utils import (
    Gatherer,
    MessageDescription,
    ParsedMeshMessage,
    ProgressCallback,
    construct_match,
)

__all__ = [
    "Model",
    "ModelConfig",
]


class Model:
    """
    Base class for mesh models.
    """

    MODEL_ID = (None, None)  # type: Tuple[Optional[int], int]
    OPCODES = []  # type: List[int]
    PUBLISH = False  # type: bool
    SUBSCRIBE = False  # type: bool

    def __init__(self, element: "Element"):
        self.__tid = 0
        self.element = element

        self.logger = self.element.logger.getChild("%s" % type(self).__name__)

        self.app_message_callbacks = defaultdict(set)  # type: Dict[int, Set[Callable]]
        self.dev_message_callbacks = defaultdict(set)  # type: Dict[int, Set[Callable]]
        self.subscription_callbacks = defaultdict(
            set
        )  # type: Dict[Union[int, UUID], Set]

        assert self.MODEL_ID[1] is not None, "A model has to have ID!"
        self.configuration = ModelConfig(bindings=[], subscriptions=[])

    def tid(self) -> int:
        tid = self.__tid
        self.__tid = (tid + 1) % 255
        return tid

    def __str__(self):
        if self.MODEL_ID[0] is None:
            return "<Model %04x>" % self.MODEL_ID[1]

        return "<VendorModel %04x%04x>" % self.MODEL_ID

    @property
    def _node_interface(self):
        return self.element.application.node_interface

    def update_configuration(self, configuration: "ModelConfig"):
        if configuration.bindings is not None:
            self.configuration.bindings = configuration.bindings

        if configuration.publication_period is not None:
            self.configuration.publication_period = configuration.publication_period

        if configuration.subscriptions is not None:
            self.configuration.subscriptions = configuration.subscriptions

        self.logger.info("Update config of %s: %s", self.MODEL_ID, self.configuration)

    def message_received(
        self,
        source: int,
        app_index: int,
        destination: Union[int, UUID],
        message: ParsedMeshMessage,
    ):
        self.logger.debug(
            "App message received: %04x [app_index %d, destination %04x] opcode=%s params=%r",
            source,
            app_index,
            destination,
            message["opcode"],
            message["params"],
        )

        callbacks = self.app_message_callbacks[message["opcode"]]

        for callback in list(callbacks):
            if callback(source, app_index, destination, message):
                callbacks.remove(callback)

        subscription_callbacks = self.subscription_callbacks.get(destination, ())

        for callback in subscription_callbacks:
            callback(
                source=source,
                destination=destination,
                app_index=app_index,
                message=message,
            )

    def dev_key_message_received(
        self, source: int, remote: bool, net_index: int, message: ParsedMeshMessage
    ):
        self.logger.debug(
            "Dev message received: %04x [remote %s, net_index %d] opcode=%s params=%r",
            source,
            remote,
            net_index,
            message["opcode"],
            message["params"],
        )

        callbacks = self.dev_message_callbacks[message["opcode"]]

        for callback in list(callbacks):
            if callback(source, net_index, message):
                callbacks.remove(callback)

    def expect_app(
        self,
        source: int,
        app_index: int,
        destination: Optional[Union[int, UUID]],
        opcode: int,
        params: MessageDescription,
    ) -> asyncio.Future:
        """
        Create an `asyncio.Future` that gets fulfilled when a specific
        application message is received.

        `opcode` is the expected message's numeric opcode, as defined in the
        mesh spec. See :py:mod:`bluetooth_mesh.messages` for a list of well-known
        opcodes.

        `params` define the message contents. After a received message is
        parsed using :py:data:`bluetooth_mesh.messages.AccessMessage`, it's compared
        recursively, key-by-key, with `params`. If all keys match, the future's
        result is set to the parsed message. If the keys are not specified,
        they will be ignored.

        One can specify wildcards by specifying key value as
        :py:data:`typing.Any` or :py:data:`...` (Ellipsis) - these match with
        anything, including composite fields.

        :param source: Sender address
        :param app_index: Index of the application key
        :param destination: Destination address
        :param opcode: Expected message opcode.
        :param params: Expected message parameters.
        """

        future = asyncio.Future()

        def app_message_received(
            _source: int,
            _app_index: int,
            _destination: Union[int, UUID],
            message: ParsedMeshMessage,
        ):
            if (_source != source) or (_app_index != app_index):
                return False

            if destination is not None and (_destination != destination):
                return False

            if params and not construct_match(message["params"], params):
                return False

            if not future.done():
                future.set_result(message)

            return True

        self.app_message_callbacks[opcode].add(app_message_received)

        return future

    def expect_dev(
        self, source: int, net_index: int, opcode: int, params: MessageDescription
    ) -> asyncio.Future:
        """
        Create an `asyncio.Future` that gets fulfilled when a specific
        device message is received.

        Unlike application keys, device keys are bound to all subnets, so
        you need to explicitly provide `net_index`.

        See :py:func:`expect_app`

        :param source: Sender address
        :param net_index: Index of the network key
        :param opcode: Expected message opcode.
        :param params: Expected message parameters.
        """

        future = asyncio.Future()

        def dev_message_received(
            _source: int, _net_index: int, message: ParsedMeshMessage
        ):
            if (_source != source) or (_net_index != net_index):
                return False

            if params and not construct_match(message["params"], params):
                return False

            if not future.done():
                future.set_result(message)

            return True

        self.dev_message_callbacks[opcode].add(dev_message_received)

        return future

    async def _send_app(self, destination: int, app_index: int, data: bytes):
        self.logger.debug(
            "Sending: %s -> %04x [app_index %d] %s",
            self.element.path,
            destination,
            app_index,
            data.hex(),
        )
        await self._node_interface.send(self.element.path, destination, app_index, data)

    async def send_app(
        self, destination: int, app_index: int, opcode: int, params: MessageDescription
    ):
        """
        Send a message using an application key.

        :param destination: Destination address
        :param app_index: Index of the application key
        :param opcode: Message opcode.
        :param params: Message parameters.
        """

        data = AccessMessage.build(dict(opcode=opcode, params=params))

        message = AccessMessage.parse(data)
        self.logger.debug(
            "Sending: %s -> %04x [app_index %d] %s %r",
            self.element.path,
            destination,
            app_index,
            message["opcode"],
            message["params"],
        )

        await self._send_app(destination, app_index, data)

    async def _send_dev(
        self, destination: int, remote: bool, net_index: int, data: bytes
    ):
        self.logger.debug(
            "Sending: %s -> %04x [remote %s, net_index %d] %s",
            self.element.path,
            destination,
            remote,
            net_index,
            data.hex(),
        )

        await self._node_interface.dev_key_send(
            self.element.path, destination, remote, net_index, data
        )

    async def send_dev(
        self, destination: int, net_index: int, opcode: int, params: MessageDescription
    ):
        """
        Send a message using a device key.

        `destination` determines which device key is going to be used. For
        local destinations, the key is always known, but if `destination` is a
        remote node, the caller must first import its device key using
        :func:`~bluetooth_mesh.application.Application.import_remote_node`.

        Unlike application keys, device keys are bound to all subnets, so
        you need to explicitly provide `net_index`.

        See :py:func:`send_app`

        :param destination: Destination address
        :param net_index: Index of the network key
        :param opcode: Message opcode.
        :param params: Message parameters.
        """

        remote = True
        data = AccessMessage.build(dict(opcode=opcode, params=params))

        message = AccessMessage.parse(data)
        self.logger.debug(
            "Sending: %s -> %04x [remote %s, net_index %d] %s %r",
            self.element.path,
            destination,
            remote,
            net_index,
            message["opcode"],
            message["params"],
        )

        await self._send_dev(destination, remote, net_index, data)

    async def repeat(
        self,
        request: Callable[[], Awaitable],
        *,
        retransmissions: int = 6,
        send_interval: float = 0.05,
    ):
        """
        Application retransmissions

        :param request:
        :param retransmissions:
        :param send_interval:

        """

        for _ in range(retransmissions):
            await request()
            await asyncio.sleep(send_interval)

    async def query(
        self,
        request: Callable[[], Awaitable[None]],
        status: asyncio.Future,
        *,
        send_interval: float = 0.2,
        timeout: float = 2.0,
    ) -> Any:
        """
        Query

        :param request:
        :param status:
        :param send_interval:
        :param timeout:

        """

        async def sender():
            while not status.done():
                await request()
                await asyncio.sleep(send_interval)

        sender = asyncio.ensure_future(sender())

        try:
            await asyncio.wait_for(status, timeout=timeout)
        finally:
            with suppress(asyncio.CancelledError):
                sender.cancel()
                await sender

        return status.result()

    async def bulk_query(
        self,
        requests: Mapping[Hashable, Callable[[], Awaitable[None]]],
        statuses: Mapping[Hashable, asyncio.Future],
        *,
        send_interval: float = 0.5,
        progress_callback: Optional[ProgressCallback] = None,
        timeout: float = 5.0,
    ) -> Mapping[Hashable, Any]:
        """
        Bulk query

        :param requests:
        :param statuses:
        :param send_interval:
        :param progress_callback:
        :param timeout:

        """

        done = {}

        async def sender():
            while requests:
                for request in list(requests.values()):
                    await request()
                    await asyncio.sleep(send_interval)

        key_mapping = {status: key for key, status in statuses.items()}

        sender = asyncio.ensure_future(sender())
        async for status, result in Gatherer(statuses.values(), timeout=timeout):
            key = key_mapping[status]
            requests.pop(key)
            done[key] = result

            if progress_callback:
                cb = progress_callback(key, result, done, statuses)
                if inspect.isawaitable(cb):
                    await cb

        with suppress(asyncio.CancelledError):
            sender.cancel()
            await sender

        results = {}

        for status in statuses.values():
            key = key_mapping[status]
            if status.cancelled():
                results[key] = asyncio.CancelledError()
            else:
                results[key] = status.exception() or status.result()

        return results

    async def subscribe(
        self,
        app_keys: Sequence[Tuple[int, int, "ApplicationKey"]],
        subscription_address: int,
        callback: Callable[[int, Union[int, UUID], int, ParsedMeshMessage], None],
    ) -> "ModelSubscriptionStatus":
        """
        Subscribe to messages sent to `subscription_address` and encrypted with
        either of application keys

        This method ensures that the application key is bound to the model.

        This causes `callback` to be called whenerver such message is received.

        :param app_keys:
        :param subscription_address:
        :param callback:

        :return: A tuple of:
            - unicast address of the element the bound model belongs to
            - subscription address (usually a group address)
            - class object of the bound model
        """
        from bluetooth_mesh.models.models import ModelSubscriptionStatus

        self.subscription_callbacks[subscription_address].add(callback)

        for app_index, *_ in app_keys:
            if app_index in self.configuration.bindings:
                continue

            await self.bind(app_index)

        if subscription_address in self.configuration.subscriptions:
            element_address = self.element.application.address + self.element.index
            return ModelSubscriptionStatus(
                element_address, subscription_address, type(self)
            )

        return await self.element.application.subscribe_model(
            subscription_address, model=self
        )

    async def unsubscribe(
        self,
        subscription_address: Optional[int] = None,
        callback: Optional[
            Callable[[int, Union[int, UUID], int, ParsedMeshMessage], None]
        ] = None,
    ) -> "ModelSubscriptionStatus":
        """
        Unubscribe from messages sent to `subscription_address`, or clear
        subscriptions if `subscription_address` is not provided.

        :return: A tuple of:
            - unicast address of the element the bound model belongs to
            - subscription address (usually a group address)
            - class object of the bound model
        """
        from bluetooth_mesh.models.models import ModelSubscriptionStatus

        if subscription_address is None:
            self.subscription_callbacks.clear()
            return await self.element.application.clear_subscriptions(model=self)

        self.subscription_callbacks[subscription_address].discard(callback)

        # if there are any callbacks left, do not unsubscribe the model
        if self.subscription_callbacks[subscription_address]:
            return

        if subscription_address not in self.configuration.subscriptions:
            element_address = self.element.application.address + self.element.index
            return ModelSubscriptionStatus(
                element_address, subscription_address, type(self)
            )

        return await self.element.application.unsubscribe_model(
            subscription_address, model=self
        )

    async def bind(self, app_key_index: int) -> "ModelBindStatus":
        """
        Bind to application key with index `app_key_index`.

        You probably don't want to call this directly, see
        :py:func:`Model.subscribe` instead.

        :param app_key_index:

        :return: A tuple of:
            - unicast address of the element the bound model belongs to
            - index of the bound application key
            - class object of the bound model
        """

        return await self.element.application.bind_app_key(app_key_index, model=self)


class ModelConfig:
    """
    Model Configuration class for mesh models.
    """

    def __init__(
        self,
        bindings: List[int] = None,
        publication_period: timedelta = None,
        subscriptions: Set[Union[int, UUID]] = None,
    ):
        self.bindings = bindings
        self.publication_period = publication_period
        self.subscriptions = subscriptions

    def __str__(self):
        return f"<ModelConfig bindings={self.bindings}, subs={self.subscriptions}>"



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import Embedded, Int8ul, Int16ul, Select, Struct, Switch, this

from bluetooth_mesh.messages.config import StatusCodeAdapter
from bluetooth_mesh.messages.generics import (
    OptionalSetParameters,
    TransitionTime,
    TransitionTimeAdapter,
)
from bluetooth_mesh.messages.util import EnumAdapter, Opcode


class LightCTLOpcode(IntEnum):
    CTL_GET = 0x825D
    CTL_SET = 0x825E
    CTL_SET_UNACKNOWLEDGED = 0x825F
    CTL_STATUS = 0x8260
    CTL_TEMPERATURE_GET = 0x8261
    CTL_TEMPERATURE_RANGE_GET = 0x8262
    CTL_TEMPERATURE_RANGE_STATUS = 0x8263
    CTL_TEMPERATURE_SET = 0x8264
    CTL_TEMPERATURE_SET_UNACKNOWLEDGED = 0x8265
    CTL_TEMPERATURE_STATUS = 0x8266
    CTL_TEMPERATURE_DEFAULT_GET = 0x8267
    CTL_TEMPERATURE_DEFAULT_STATUS = 0x8268

    def __repr__(self):
        return str(self.value)


class LightCTLSetupOpcode(IntEnum):
    CTL_TEMPERATURE_DEFAULT_SET = 0x8269
    CTL_TEMPERATURE_DEFAULT_SET_UNACKNOWLEDGED = 0x826A
    CTL_TEMPERATURE_RANGE_SET = 0x826B
    CTL_TEMPERATURE_RANGE_SET_UNACKNOWLEDGED = 0x826C

    def __repr__(self):
        return str(self.value)


# fmt: off
LightCTLGet = Struct()

LightCTLDefault = Struct(
    "ctl_lightness" / Int16ul,
    "ctl_temperature" / Int16ul,
    "ctl_delta_uv" / Int16ul
)

LightCTLSetMinimal = Struct(
    Embedded(LightCTLDefault),
    "tid" / Int8ul
)

LightCTLSetOptional = Struct(
    Embedded(LightCTLSetMinimal),
    Embedded(OptionalSetParameters)
)

LightCTLSet = Select(
    LightCTLSetOptional,
    LightCTLSetMinimal
)

LightCTLStatusMinimal = Struct(
    "present_ctl_lightness" / Int16ul,
    "present_ctl_temperature" / Int16ul,
)

LightCTLStatusOptional = Struct(
    Embedded(LightCTLStatusMinimal),
    "target_ctl_lightness" / Int16ul,
    "target_ctl_temperature" / Int16ul,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=True)
)

LightCTLStatus = Select(
    LightCTLStatusOptional,
    LightCTLStatusMinimal
)

LightCTLTemperatureStatusMinimal = Struct(
    "present_ctl_temperature" / Int16ul,
    "present_ctl_delta_uv" / Int16ul,
)

LightCTLTemperatureStatusOptional = Struct(
    Embedded(LightCTLTemperatureStatusMinimal),
    "target_ctl_temperature" / Int16ul,
    "target_ctl_delta_uv" / Int16ul,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=True)
)

LightCTLTemperatureStatus = Select(
    LightCTLTemperatureStatusOptional,
    LightCTLTemperatureStatusMinimal
)

LightCTLTemperatureSetMinimal = Struct(
    "ctl_temperature" / Int16ul,
    "ctl_delta_uv" / Int16ul,
    "tid" / Int8ul
)

LightCTLTemperatureSetOptional = Struct(
    Embedded(LightCTLTemperatureSetMinimal),
    Embedded(OptionalSetParameters)
)

LightCTLTemperatureSet = Select(
    LightCTLTemperatureSetOptional,
    LightCTLTemperatureSetMinimal
)

LightCTLRange = Struct(
    "range_min" / Int16ul,
    "range_max" / Int16ul,
)

LightCTLRangeStatus = Struct(
    "status" / StatusCodeAdapter,
    Embedded(LightCTLRange)
)


LightCTLMessage = Struct(
    "opcode" / Opcode(LightCTLOpcode),
    "params" / Switch(
        this.opcode,
        {
            LightCTLOpcode.CTL_GET: LightCTLGet,
            LightCTLOpcode.CTL_SET: LightCTLSet,
            LightCTLOpcode.CTL_SET_UNACKNOWLEDGED: LightCTLSet,
            LightCTLOpcode.CTL_STATUS: LightCTLStatus,
            LightCTLOpcode.CTL_TEMPERATURE_GET: LightCTLGet,
            LightCTLOpcode.CTL_TEMPERATURE_RANGE_GET: LightCTLGet,
            LightCTLOpcode.CTL_TEMPERATURE_RANGE_STATUS: LightCTLRangeStatus,
            LightCTLOpcode.CTL_TEMPERATURE_SET: LightCTLTemperatureSet,
            LightCTLOpcode.CTL_TEMPERATURE_SET_UNACKNOWLEDGED: LightCTLTemperatureSet,
            LightCTLOpcode.CTL_TEMPERATURE_STATUS: LightCTLTemperatureStatus,
            LightCTLOpcode.CTL_TEMPERATURE_DEFAULT_GET: LightCTLGet,
            LightCTLOpcode.CTL_TEMPERATURE_DEFAULT_STATUS: LightCTLDefault,
        },
    )
)


LightCTLSetupMessage = Struct(
    "opcode" / Opcode(LightCTLSetupOpcode),
    "params" / Switch(
        this.opcode,
        {
            LightCTLSetupOpcode.CTL_TEMPERATURE_DEFAULT_SET: LightCTLDefault,
            LightCTLSetupOpcode.CTL_TEMPERATURE_DEFAULT_SET_UNACKNOWLEDGED: LightCTLDefault,
            LightCTLSetupOpcode.CTL_TEMPERATURE_RANGE_SET: LightCTLRange,
            LightCTLSetupOpcode.CTL_TEMPERATURE_RANGE_SET_UNACKNOWLEDGED: LightCTLRange,
        },
    )
)
# fmt: on
